"""This is a copy of the DigestAuth logic from httpx auth module with additions for cached auth.

Base copied from: https://github.com/encode/httpx/blob/32d37cfdf1697f0dbacbc2233ee34606369be87b/httpx/_auth.py
Issue tracking the need for the copy: https://github.com/encode/httpx/issues/1467
"""

import hashlib
import os
import re
import time
import typing
from urllib.request import parse_http_list

from httpx import ProtocolError
from httpx import Request, Response
from httpx import Auth


def to_bytes(value: typing.Union[str, bytes], encoding: str = "utf-8") -> bytes:
    return value.encode(encoding) if isinstance(value, str) else value

def to_str(value: typing.Union[str, bytes], encoding: str = "utf-8") -> str:
    return value if isinstance(value, str) else value.decode(encoding)

def unquote(value: str) -> str:
    return value[1:-1] if value[0] == value[-1] == '"' else value

class CachedDigestAuth(Auth):
    """Digest authentication with a cached challange response.
    
    Known issues:
     * Only a single challenge is every cached, no checks is done for hostname being correct
     * Nonce counting is not implemented since philips TV's don't seem to require it.
    """

    _ALGORITHM_TO_HASH_FUNCTION: typing.Dict[str, typing.Callable] = {
        "MD5": hashlib.md5,
        "MD5-SESS": hashlib.md5,
        "SHA": hashlib.sha1,
        "SHA-SESS": hashlib.sha1,
        "SHA-256": hashlib.sha256,
        "SHA-256-SESS": hashlib.sha256,
        "SHA-512": hashlib.sha512,
        "SHA-512-SESS": hashlib.sha512,
    }

    def __init__(
        self, username: typing.Union[str, bytes], password: typing.Union[str, bytes]
    ) -> None:
        self._challenge = None
        self._username = to_bytes(username)
        self._password = to_bytes(password)

    def auth_flow(self, request: Request) -> typing.Generator[Request, Response, None]:
        if self._challenge:
            request.headers["Authorization"] = self._build_auth_header(request, self._challenge)

        response = yield request

        if response.status_code != 401 or "www-authenticate" not in response.headers:
            # If the response is not a 401 then we don't
            # need to build an authenticated request.
            return

        for auth_header in response.headers.get_list("www-authenticate"):
            if auth_header.lower().startswith("digest "):
                break
        else:
            # If the response does not include a 'WWW-Authenticate: Digest ...'
            # header, then we don't need to build an authenticated request.
            return

        self._challenge = self._parse_challenge(request, response, auth_header)
        request.headers["Authorization"] = self._build_auth_header(request, self._challenge)
        yield request

    def _parse_challenge(
        self, request: Request, response: Response, auth_header: str
    ) -> "_DigestAuthChallenge":
        """
        Returns a challenge from a Digest WWW-Authenticate header.
        These take the form of:
        `Digest realm="realm@host.com",qop="auth,auth-int",nonce="abc",opaque="xyz"`
        """
        scheme, _, fields = auth_header.partition(" ")

        # This method should only ever have been called with a Digest auth header.
        assert scheme.lower() == "digest"

        header_dict: typing.Dict[str, str] = {}
        for field in parse_http_list(fields):
            key, value = field.strip().split("=", 1)
            header_dict[key] = unquote(value)

        try:
            realm = header_dict["realm"].encode()
            nonce = header_dict["nonce"].encode()
            algorithm = header_dict.get("algorithm", "MD5")
            opaque = header_dict["opaque"].encode() if "opaque" in header_dict else None
            qop = header_dict["qop"].encode() if "qop" in header_dict else None
            return _DigestAuthChallenge(
                realm=realm, nonce=nonce, algorithm=algorithm, opaque=opaque, qop=qop
            )
        except KeyError as exc:
            message = "Malformed Digest WWW-Authenticate header"
            raise ProtocolError(message, request=request) from exc

    def _build_auth_header(
        self, request: Request, challenge: "_DigestAuthChallenge"
    ) -> str:
        hash_func = self._ALGORITHM_TO_HASH_FUNCTION[challenge.algorithm]

        def digest(data: bytes) -> bytes:
            return hash_func(data).hexdigest().encode()

        A1 = b":".join((self._username, challenge.realm, self._password))

        path = request.url.raw_path
        A2 = b":".join((request.method.encode(), path))
        # TODO: implement auth-int
        HA2 = digest(A2)

        nonce_count = 1  # TODO: implement nonce counting
        nc_value = b"%08x" % nonce_count
        cnonce = self._get_client_nonce(nonce_count, challenge.nonce)

        HA1 = digest(A1)
        if challenge.algorithm.lower().endswith("-sess"):
            HA1 = digest(b":".join((HA1, challenge.nonce, cnonce)))

        qop = self._resolve_qop(challenge.qop, request=request)
        if qop is None:
            digest_data = [HA1, challenge.nonce, HA2]
        else:
            digest_data = [challenge.nonce, nc_value, cnonce, qop, HA2]
        key_digest = b":".join(digest_data)

        format_args = {
            "username": self._username,
            "realm": challenge.realm,
            "nonce": challenge.nonce,
            "uri": path,
            "response": digest(b":".join((HA1, key_digest))),
            "algorithm": challenge.algorithm.encode(),
        }
        if challenge.opaque:
            format_args["opaque"] = challenge.opaque
        if qop:
            format_args["qop"] = b"auth"
            format_args["nc"] = nc_value
            format_args["cnonce"] = cnonce

        return "Digest " + self._get_header_value(format_args)

    def _get_client_nonce(self, nonce_count: int, nonce: bytes) -> bytes:
        s = str(nonce_count).encode()
        s += nonce
        s += time.ctime().encode()
        s += os.urandom(8)

        return hashlib.sha1(s).hexdigest()[:16].encode()

    def _get_header_value(self, header_fields: typing.Dict[str, bytes]) -> str:
        NON_QUOTED_FIELDS = ("algorithm", "qop", "nc")
        QUOTED_TEMPLATE = '{}="{}"'
        NON_QUOTED_TEMPLATE = "{}={}"

        header_value = ""
        for i, (field, value) in enumerate(header_fields.items()):
            if i > 0:
                header_value += ", "
            template = (
                QUOTED_TEMPLATE
                if field not in NON_QUOTED_FIELDS
                else NON_QUOTED_TEMPLATE
            )
            header_value += template.format(field, to_str(value))

        return header_value

    def _resolve_qop(
        self, qop: typing.Optional[bytes], request: Request
    ) -> typing.Optional[bytes]:
        if qop is None:
            return None
        qops = re.split(b", ?", qop)
        if b"auth" in qops:
            return b"auth"

        if qops == [b"auth-int"]:
            raise NotImplementedError("Digest auth-int support is not yet implemented")

        message = f'Unexpected qop value "{qop!r}" in digest auth'
        raise ProtocolError(message, request=request)


class _DigestAuthChallenge(typing.NamedTuple):
    realm: bytes
    nonce: bytes
    algorithm: str
    opaque: typing.Optional[bytes]
    qop: typing.Optional[bytes]


from typing import Dict, List, Literal, NamedTuple, Optional, Type, TypedDict, Union


class ActivitiesChannelType(TypedDict, total=False):
    ccid: int
    name: str
    preset: str
    onid: str
    tsid: str
    sid: str
    serviceType: str
    type: str
    logoVersion: str


class ActivitiesChannelListType(TypedDict, total=False):
    id: str
    version: Union[int, str]


class ActivitesTVType(TypedDict, total=False):
    channel: ActivitiesChannelType
    channelList: ActivitiesChannelListType


class ChannelsCurrentType(TypedDict):
    id: str


class ComponentType(TypedDict):
    packageName: str
    className: str


class ApplicationIntentType(TypedDict, total=False):
    extras: Dict
    component: ComponentType
    action: str


class ApplicationType(TypedDict, total=False):
    label: str
    intent: ApplicationIntentType
    order: int
    id: str
    type: str


class ApplicationsType(TypedDict):
    version: int
    applications: List[ApplicationType]


class FavoriteChannelType(TypedDict, total=False):
    ccid: int
    preset: str


class FavoriteListType(TypedDict, total=False):
    id: str
    version: Union[int, str]
    listType: str
    medium: str
    name: str
    channels: List[FavoriteChannelType]


class ChannelType(TypedDict, total=False):
    ccid: Union[int, str]
    preset: str
    name: str
    onid: int
    tsid: int
    sid: int
    serviceType: str
    type: str
    logoVersion: Union[int, str]


class ChannelListType(TypedDict, total=False):
    id: str
    version: Union[int, str]
    listType: str
    medium: str
    operator: str
    installCountry: str
    Channel: List[ChannelType]


ChannelsType = Dict[str, ChannelType]


class ChannelDbTvListBase(TypedDict):
    id: str
    version: Union[int, str]
    listType: str


class ChannelDbTvList(ChannelDbTvListBase, total=False):
    medium: str
    active: bool
    virtual: bool
    modifiable: bool

class ChannelDbTvListFavorite(ChannelDbTvListBase, total=False):
    medium: str
    active: bool
    virtual: bool
    modifiable: bool
    parentId: str
    name: str

class ChannelDbTv(TypedDict, total=False):
    channelLists: List[ChannelDbTvList]
    favoriteLists: List[ChannelDbTvListFavorite]


class JsonFeaturesType(TypedDict, total=False):
    editfavorites: List[str]
    recordings: List[str]
    ambilight: List[str]
    menuitems: List[str]
    textentry: List[str]
    applications: List[str]
    pointer: List[str]
    inputkey: List[str]
    activities: List[str]
    channels: List[str]
    mappings: List[str]


class SystemFeaturesType(TypedDict, total=False):
    tvtype: str
    content: List[str]
    tvsearch: str
    pairing_type: str
    secured_transport: str
    companion_screen: Union[str, bool]

    # stored here on saphi
    os_type: str


class SystemFeaturingType(TypedDict):
    jsonfeatures: JsonFeaturesType
    systemfeatures: SystemFeaturesType


class SystemType(TypedDict, total=False):
    menulanguage: str
    name: str
    country: str
    serialnumber: str
    serialnumber_encrypted: str
    softwareversion: str
    softwareversion_encrypted: str
    model: str
    model_encrypted: str
    deviceid_encrypted: str
    nettvversion: str
    epgsource: str
    api_version: Dict
    featuring: SystemFeaturingType
    notifyChange: str

    # here on android systems
    os_type: str


class SourceType(TypedDict, total=False):
    name: str


SourcesType = Dict[str, SourceType]


class SourceCurrentType(TypedDict):
    id: str


class ContextType(TypedDict, total=False):
    data: str
    level1: str
    level2: str
    level3: str

    # in some older context these exists
    activity: str
    menu: str
    Recording: str


AmbilightPixelType = Dict[Literal["r", "g", "b"], int]

AmbilightSideType = Dict[str, AmbilightPixelType]

AmbilightLayerType = Dict[Literal["left", "top", "right", "bottom"], AmbilightSideType]

AmbilightLayersType = Dict[str, AmbilightLayerType]


class AmbilightSupportedStyleType(TypedDict, total=False):
    styleName: str
    algorithms: List[str]
    maxTuning: int
    maxSpeed: int

    # Extension to get the non expert modes too
    menuSettings: List[str]


class AmbilightSupportedStylesType(TypedDict):
    supportedStyles: List[AmbilightSupportedStyleType]


class AmbilightColorType(TypedDict):
    hue: int
    saturation: int
    brightness: int


class AmbilightAudioSettingsType(TypedDict):
    color: AmbilightColorType
    colorDelta: AmbilightColorType
    tuning: int
    algorithm: str


class AmbilightColorSettingsType(TypedDict):
    color: AmbilightColorType
    colorDelta: AmbilightColorType
    speed: int
    algorithm: str


class AmbilightCurrentConfiguration(TypedDict, total=False):
    styleName: str
    isExpert: bool
    menuSetting: str
    stringValue: str

    audioSettings: AmbilightAudioSettingsType
    colorSettings: AmbilightColorSettingsType

class RecordingsListedType(TypedDict, total=False):
    RecordingId: int
    ServerRecordingId: int
    RecordingType: str #RECORDING_ONGOING, RECORDING_SCHEDULED, RECORDING_NEW, RECORDING_PARTIALLY_VIEWED, RECORDING_VIEWED
    RecName: str
    EventId: int
    EventInfo: str
    EventExtendedInfo: str
    EventGenre: str
    EventRating: int
    SeriesID: str
    StartTime: int
    ActualStartTime: int
    Duration: int
    MarginStart: int
    MarginEnd: int
    AutoMarginStart: int
    AutoMarginEnd: int
    RetentionInfo: int
    SeasonNo: int
    EpisodeNo: int
    EpisodeCount: int
    ProgramNumber: int
    ProgramDuration: int
    ResumeInfo: int
    LastPinChangedTime: int
    HasCicamPin: bool
    hasDot: bool
    HasLicenseFile: bool
    IsRadio: bool
    IsPartial: bool
    IsIpEpgRec: bool
    isFTARecording: bool
    EITVersion: int
    EITSource: str #EIT_SOURCE_PF, ...
    RecError: str #REC_ERROR_NONE, ...
    Version: int
    Size: int
    ccid: int

class RecordingsListed(TypedDict):
    version: str
    recordings: List[RecordingsListedType]

class MenuItemsSettingsNodeDataSliderData(TypedDict):
    min: int
    max: int
    step_size: int

class MenuItemsSettingsNodeDataSlider(TypedDict):
    slider_id: str
    slider_data: MenuItemsSettingsNodeDataSliderData

class MenuItemsSettingsNodeDataEnumRequired(TypedDict):
    enum_id: int
    string_id: str

class MenuItemsSettingsNodeDataEnum(MenuItemsSettingsNodeDataEnumRequired, total=False):
    icon: Optional[str]

class MenuItemsSettingsNodeData(TypedDict, total=False):
    nodes: List["MenuItemsSettingsNode"]              # PARENT_NODE
    enums: List[MenuItemsSettingsNodeDataEnum]        # LIST_NODE
    slider_data: MenuItemsSettingsNodeDataSliderData  # SLIDER_NODE
    sliders: List[MenuItemsSettingsNodeDataSlider]    # MULTIPLE_SLIDERS
    colors: List[int]                                 # WALL_COLOR_NODE

class MenuItemsSettingsNodeRequired(TypedDict):
    node_id: int
    type: str
    data: MenuItemsSettingsNodeData

class MenuItemsSettingsNode(MenuItemsSettingsNodeRequired, total=False):
    string_id: str
    icon: Optional[str]
    context: str

class MenuItemsSettingsEntry(NamedTuple):
    node: MenuItemsSettingsNode
    parent: Optional[int]

class MenuItemsSettingsStructure(TypedDict, total=False):
    node: MenuItemsSettingsNode

class MenuItemsSettingsValueInt(TypedDict):
    value: int

class MenuItemsSettingsValueBool(TypedDict):
    value: bool

class MenuItemsSettingsValueEnumEntry(TypedDict):
    enum_id: int
    controllable: bool
    available: bool
    string_id: str

class MenuItemsSettingsValueEnumRequired(TypedDict):
    select_item: Optional[int]

class MenuItemsSettingsValueEnum(MenuItemsSettingsValueEnumRequired, total=False):
    enum_values: List[MenuItemsSettingsValueEnumEntry]

class MenuItemsSettingsValueNode(TypedDict):
    activenode_id: int

class MenuItemsSettingsValueSlider(TypedDict):
    slider_id: int
    value: int

class MenuItemsSettingsValueSliders(TypedDict):
    values: List[MenuItemsSettingsValueSlider]

MenuItemsSettingsValueData = Union[
    MenuItemsSettingsValueInt,
    MenuItemsSettingsValueBool,
    MenuItemsSettingsValueEnum,
    MenuItemsSettingsValueNode,
    MenuItemsSettingsValueSliders
]

class MenuItemsSettingsCurrentValueValue(TypedDict):
    Nodeid: int
    Controllable: bool
    Available: bool
    string_id: str
    data: MenuItemsSettingsValueData

class MenuItemsSettingsCurrentValue(TypedDict):
    value: MenuItemsSettingsCurrentValueValue

class MenuItemsSettingsCurrent(TypedDict):
    values: List[MenuItemsSettingsCurrentValue]
    version: int

class MenuItemsSettingsCurrentPostNode(TypedDict):
    nodeid: int

class MenuItemsSettingsCurrentPost(TypedDict):
    nodes: List[MenuItemsSettingsCurrentPostNode]

class MenuItemsSettingsUpdateValueValue(TypedDict):
    Nodeid: int
    data: MenuItemsSettingsValueData

class MenuItemsSettingsUpdateValue(TypedDict):
    value: MenuItemsSettingsUpdateValueValue

class MenuItemsSettingsUpdate(TypedDict):
    values: List[MenuItemsSettingsUpdateValue]

class StringsTranslation(TypedDict):
    string_id: str
    string_translation: Optional[str]

class Strings(TypedDict):
    translations: List[StringsTranslation]

class StringId(TypedDict):
    string_id: str

class StringLocale(TypedDict):
    language: str
    country: str
    variant: str

class StringsRequest(TypedDict):
    strings: List[StringId]
    locale: StringLocale



from aiohttp import web
from aiohttp.web_request import Request
from aiohttp.web_response import json_response

from haphilipsjs.data.v1 import (
    AMBILIGHT,
    CHANNELLISTS,
    CHANNELS_CURRENT,
    CHANNELS,
    SOURCES_CURRENT,
    SOURCES,
    SYSTEM_55PFL6007T,
    VOLUME,
)


async def put_ambilight_cached(req):
    print("cached")
    return web.json_response({})


async def put_ambilight_mode(req):
    print("mode")
    return web.json_response({})


def get_data(data):
    async def get(req: Request):
        return web.json_response(data)

    return get


app = web.Application()
app.add_routes(
    [
        web.get("/1/system", get_data(SYSTEM_55PFL6007T)),
        web.get("/1/sources", get_data(SOURCES)),
        web.get("/1/sources/current", get_data(SOURCES_CURRENT)),
        web.get("/1/channels", get_data(CHANNELS)),
        web.get("/1/channels/current", get_data(CHANNELS_CURRENT)),
        web.get("/1/audio/volume", get_data(VOLUME)),
        web.get("/1/channellists", get_data(CHANNELLISTS)),
        web.get("/1/ambilight/mode", get_data(AMBILIGHT["mode"])),
        web.post("/1/ambilight/mode", put_ambilight_mode),
        web.get("/1/ambilight/topology", get_data(AMBILIGHT["topology"])),
        web.get("/1/ambilight/measured", get_data(AMBILIGHT["measured"])),
        web.get("/1/ambilight/processed", get_data(AMBILIGHT["processed"])),
        web.get("/1/ambilight/cached", get_data(AMBILIGHT["cached"])),
        web.post("/1/ambilight/cached", put_ambilight_cached),
    ]
)

if __name__ == "__main__":
    web.run_app(app, port=1925)



import httpx
import haphilipsjs
import pytest
import respx
import json
from typing import NamedTuple, cast, Dict

from haphilipsjs.data.v1 import (
    AMBILIGHT,
    CHANNELLISTS,
    CHANNELS_CURRENT,
    CHANNELS,
    CONTEXT,
    POWERSTATE,
    SOURCES_CURRENT,
    SOURCES,
    SYSTEM_DECRYPTED,
    SYSTEM_ENCRYPTED,
    VOLUME,
)


class Param(NamedTuple):
    type: str
    base: str


@pytest.fixture(params=["55PFL6007T", "47PFH6309", "65PUS6121"], name="param")
async def param_fixture(request):
    return Param(request.param, "http://127.0.0.1:1925/1")


@pytest.fixture
async def client_mock(param: Param):
    with respx.mock:
        client = haphilipsjs.PhilipsTV("127.0.0.1", api_version=1)

        respx.get(f"{param.base}/system").respond(
            json=cast(Dict, SYSTEM_ENCRYPTED[param.type])
        )
        respx.get(f"{param.base}/sources").respond(json=SOURCES)
        respx.get(f"{param.base}/sources/current").respond(
            json=cast(Dict, SOURCES_CURRENT)
        )
        respx.get(f"{param.base}/channels").respond(json=CHANNELS)
        respx.get(f"{param.base}/channels/current").respond(
            json=cast(Dict, CHANNELS_CURRENT)
        )
        respx.get(f"{param.base}/audio/volume").respond(json=VOLUME)
        respx.get(f"{param.base}/channellists").respond(json=CHANNELLISTS)
        respx.get(f"{param.base}/ambilight/mode").respond(json=AMBILIGHT["mode"])
        respx.get(f"{param.base}/ambilight/topology").respond(
            json=AMBILIGHT["topology"]
        )
        respx.get(f"{param.base}/ambilight/measured").respond(
            json=AMBILIGHT["measured"]
        )
        respx.get(f"{param.base}/ambilight/processed").respond(
            json=AMBILIGHT["processed"]
        )
        respx.get(f"{param.base}/ambilight/cached").respond(json=AMBILIGHT["cached"])
        if param.type in POWERSTATE:
            respx.get(f"{param.base}/powerstate").respond(json=POWERSTATE[param.type])
        else:
            respx.get(f"{param.base}/powerstate").respond(404)
        if param.type in CONTEXT:
            respx.get(f"{param.base}/context").respond(json=CONTEXT[param.type])
        else:
            respx.get(f"{param.base}/context").respond(404)

        yield client
        await client.aclose()


async def test_basic_data(client_mock: haphilipsjs.PhilipsTV, param: Param):
    """Test for basic data"""
    await client_mock.update()
    assert client_mock.on == True
    assert client_mock.system == SYSTEM_DECRYPTED[param.type]
    assert client_mock.sources == SOURCES
    assert client_mock.channels == CHANNELS
    assert client_mock.channels_current == [
        {
            **channel,
            "ccid": key
        }
        for key, channel in CHANNELS.items()
    ]
    assert client_mock.ambilight_current_configuration is None
    assert client_mock.ambilight_styles == {}
    assert client_mock.powerstate == POWERSTATE.get(param.type, {}).get("powerstate")
    assert client_mock.context == CONTEXT.get(param.type)


async def test_current_source_none(client_mock, param: Param):
    await client_mock.update()
    assert client_mock.source_id == "hdmi1"

    respx.get(f"{param.base}/sources/current").respond(json={})
    await client_mock.update()
    assert client_mock.source_id == None


async def test_current_channel_none(client_mock, param: Param):
    await client_mock.update()
    assert client_mock.channel_id == "fingerprint-1"

    respx.get(f"{param.base}/channels/current").respond(json={})
    await client_mock.update()
    assert client_mock.channel_id == None


async def test_current_channel_with_channellist_prefix(client_mock, param: Param):
    respx.get(f"{param.base}/channels/current").respond(json={"id": "0_fingerprint-1"})
    await client_mock.update()
    assert client_mock.channel_id == "fingerprint-1"


async def test_get_source_name(client_mock):
    """Verify that we can translate source id to name"""
    await client_mock.update()
    assert await client_mock.getSourceName("ypbpr") == "Y Pb Pr"
    assert await client_mock.getSourceName("invalid_name") == None


async def test_get_channel_name(client_mock):
    """Verify that we can translate channel id to name"""
    await client_mock.update()
    assert await client_mock.getChannelName("fingerprint-3") == "Irdeto scrambled"
    assert await client_mock.getChannelName("invalid_name") == None


async def test_timeout(client_mock, param: Param):
    """Test that connect timeouts trigger tv to be considered off"""
    await client_mock.update()
    assert client_mock.on == True
    respx.get(f"{param.base}/sources/current").mock(side_effect=httpx.ConnectTimeout)

    await client_mock.update()
    assert client_mock.on == False

    respx.get(f"{param.base}/sources/current").respond(json=cast(Dict, SOURCES_CURRENT))

    await client_mock.update()
    assert client_mock.on == True


async def test_volume(client_mock, param: Param):
    respx.get(f"{param.base}/audio/volume").respond(
        json={"muted": True, "current": 30, "min": 0, "max": 60}
    )

    await client_mock.update()
    assert client_mock.volume == 0.5
    assert client_mock.muted == True

    respx.get(f"{param.base}/audio/volume").respond(
        json={"muted": False, "current": 60, "min": 0, "max": 60}
    )

    await client_mock.update()
    assert client_mock.volume == 1.0
    assert client_mock.muted == False

    respx.get(f"{param.base}/audio/volume").respond(
        json={"muted": False, "current": 0, "min": 0, "max": 60}
    )

    await client_mock.update()
    assert client_mock.volume == 0.0
    assert client_mock.muted == False


async def test_set_volume(client_mock, param: Param):
    await client_mock.update()
    respx.post(f"{param.base}/audio/volume").respond(json={})

    await client_mock.setVolume(0.5)

    assert respx.calls[-1].request.url == f"{param.base}/audio/volume"
    assert json.loads(respx.calls[-1].request.content) == {
        "muted": False,
        "current": 30,
    }

    await client_mock.setVolume(1.0, True)

    assert json.loads(respx.calls[-1].request.content) == {
        "muted": True,
        "current": 60,
    }

    await client_mock.setVolume(0.0, True)

    assert json.loads(respx.calls[-1].request.content) == {
        "muted": True,
        "current": 0,
    }

    await client_mock.setVolume(None, True)

    assert json.loads(respx.calls[-1].request.content) == {
        "muted": True,
    }


async def test_set_source(client_mock, param: Param):
    respx.post(f"{param.base}/sources/current").respond(json={})

    await client_mock.setSource("hdmi2")

    assert respx.calls[-1].request.url == f"{param.base}/sources/current"
    assert json.loads(respx.calls[-1].request.content) == {
        "id": "hdmi2",
    }


async def test_set_channel(client_mock, param: Param):
    respx.post(f"{param.base}/channels/current").respond(json={})

    await client_mock.setChannel("fingerprint-2")

    assert respx.calls[-1].request.url == f"{param.base}/channels/current"
    assert json.loads(respx.calls[-1].request.content) == {
        "id": "fingerprint-2",
    }


async def test_send_key(client_mock, param: Param):
    respx.post(f"{param.base}/input/key").respond(json={})

    await client_mock.sendKey("Standby")

    assert respx.calls[-1].request.url == f"{param.base}/input/key"
    assert json.loads(respx.calls[-1].request.content) == {
        "key": "Standby",
    }


async def test_send_key_off(client_mock, param: Param):
    respx.post(f"{param.base}/input/key").mock(side_effect=httpx.ConnectTimeout)

    with pytest.raises(haphilipsjs.ConnectionFailure):
        await client_mock.sendKey("Standby")


async def test_ambilight_mode(client_mock, param: Param):
    await client_mock.getSystem()

    respx.post(f"{param.base}/ambilight/mode").respond(json={})
    await client_mock.setAmbilightMode("internal")

    assert respx.calls[-1].request.url == f"{param.base}/ambilight/mode"
    assert json.loads(respx.calls[-1].request.content) == {
        "current": "internal",
    }

    assert await client_mock.getAmbilightMode()
    assert client_mock.ambilight_mode == "internal"

    assert await client_mock.setAmbilightMode("manual")
    assert client_mock.ambilight_mode == "manual"


async def test_ambilight_power(client_mock: haphilipsjs.PhilipsTV, param: Param):
    respx.post(f"{param.base}/ambilight/cached").respond(json={})
    respx.post(f"{param.base}/ambilight/mode").respond(json={})

    assert client_mock.ambilight_power == None
    await client_mock.update()
    await client_mock.getAmbilightPower()
    await client_mock.getAmbilightCached()
    await client_mock.getAmbilightMode()

    assert client_mock.ambilight_power == "On"

    assert await client_mock.setAmbilightPower("Off")

    assert respx.calls[-2].request.url == f"{param.base}/ambilight/cached"
    assert json.loads(respx.calls[-2].request.content) == {"r": 0, "g": 0, "b": 0}

    assert respx.calls[-1].request.url == f"{param.base}/ambilight/cached"

    assert await client_mock.setAmbilightPower("On")

    assert respx.calls[-1].request.url == f"{param.base}/ambilight/mode"
    assert json.loads(respx.calls[-1].request.content) == {"current": "internal"}


async def test_ambilight_topology(client_mock):
    assert await client_mock.getAmbilightTopology() == AMBILIGHT["topology"]


async def test_ambilight_measured(client_mock):
    assert await client_mock.getAmbilightMeasured() == AMBILIGHT["measured"]


async def test_ambilight_processed(client_mock):
    assert await client_mock.getAmbilightProcessed() == AMBILIGHT["processed"]


async def test_ambilight_cached(client_mock, param: Param):
    assert await client_mock.getAmbilightCached() == AMBILIGHT["cached"]

    respx.post(f"{param.base}/ambilight/cached").respond(json={})

    data = {"r": 100, "g": 210, "b": 30}

    await client_mock.setAmbilightCached(data)

    assert respx.calls[-2].request.url == f"{param.base}/ambilight/cached"

    assert respx.calls[-1].request.url == f"{param.base}/ambilight/cached"
    assert json.loads(respx.calls[-2].request.content) == data



from typing import cast
from haphilipsjs.typing import (
    ChannelsCurrentType,
    SourcesType,
    SystemType,
    SourceCurrentType,
    ChannelsType,
)


SYSTEM_55PFL6007T: SystemType = {
    "menulanguage": "English",
    "name": "Philips TV",
    "country": "Sweden",
    "serialnumber": "1234567890",
    "softwareversion": "abcd",
    "model": "55PFL6007T/12",
}

SYSTEM_47PFH6309: SystemType = {
    "menulanguage": "Spanish",
    "name": "PHILIPS TV",
    "country": "Spain",
    "nettvversion": "",
    "epgsource": "broadcast",
}

SYSTEM_47PFH6309_ENCRYPTED = cast(
    SystemType,
    {
        **SYSTEM_47PFH6309,
        "serialnumber_encrypted": "bf1BcncGiQyBVS47ZXFWjNXoynlKUNlqDhxQz5ikPEU=",
        "softwareversion_encrypted": "K2kseVsmQFgkd15gKkJ+ao+IN30u9WA8apvZ8LuQYkrUGEcxIhN1tkDM7wnn3V+5",
        "model_encrypted": "K2kseVsmQFgkd15gKkJ+akRXrm2rweXcLkCrwpCecFY=",
        "deviceid_encrypted": "",
    },
)

SYSTEM_47PFH6309_DECRYPTED = cast(
    SystemType,
    {
        **SYSTEM_47PFH6309,
        "serialnumber": "ABCDEFGHIJKLF",
        "softwareversion": "mt5880:TPN141E_010.003.086.128",
        "model": "1_47PFH6309/88",
        "deviceid": "",
    },
)

SYSTEM_65PUS6121: SystemType = {
    "menulanguage": "Dutch",
    "name": "65PUS6121",
    "country": "Netherlands",
    "nettvversion": "4.5.0",
    "epgsource": "broadcast",
    "featuring": {
        "jsonfeatures": {
            "recordings": ["List", "Schedule", "Manage"],
            "ambilight": ["LoungeLight"],
            "textentry": ["context_based", "initial_string_available"],
            "inputkey": ["key", "unicode"],
            "pointer": ["context_based"],
            "activities": ["browser"],
        },
        "systemfeatures": {"tvtype": "consumer", "content": ["dmr", "dms_tad"]},
    },
}

SYSTEM_65PUS6121_ENCRYPTED = cast(
    SystemType,
    {
        **SYSTEM_65PUS6121,
        "serialnumber_encrypted": "bf1BcncGiQyBVS47ZXFWjNXoynlKUNlqDhxQz5ikPEU=",
        "softwareversion_encrypted": "K2kseVsmQFgkd15gKkJ+avGAg/2Fns000yNVcCibCqJc8lV5YE2qHdE/pFg2ib2M",
        "model_encrypted": "K2kseVsmQFgkd15gKkJ+ai592Vkjcax6Xprb+fzv08s=",
        "deviceid_encrypted": "K2kseVsmQFgkd15gKkJ+arY7DmesJhIol3fe863YYY0=",
    },
)

SYSTEM_65PUS6121_DECRYPTED = cast(
    SystemType,
    {
        **SYSTEM_65PUS6121,
        "serialnumber": "ABCDEFGHIJKLF",
        "softwareversion": "mt5891:TPL161E_012.003.039.001",
        "model": "1_65PUS6121/12",
        "deviceid": "54885869",
    },
)

SYSTEM_ENCRYPTED = {
    "55PFL6007T": SYSTEM_55PFL6007T,
    "47PFH6309": SYSTEM_47PFH6309_ENCRYPTED,
    "65PUS6121": SYSTEM_65PUS6121_ENCRYPTED,
}

SYSTEM_DECRYPTED = {
    "55PFL6007T": SYSTEM_55PFL6007T,
    "47PFH6309": SYSTEM_47PFH6309_DECRYPTED,
    "65PUS6121": SYSTEM_65PUS6121_DECRYPTED,
}

SOURCES: SourcesType = {
    "tv": {"name": "Watch TV"},
    "satellite": {"name": "Watch satellite"},
    "hdmi1": {"name": "HDMI 1"},
    "hdmi2": {"name": "HDMI 2"},
    "hdmi3": {"name": "HDMI 3"},
    "hdmiside": {"name": "HDMI side"},
    "ext1": {"name": "EXT 1"},
    "ext2": {"name": "EXT 2"},
    "ypbpr": {"name": "Y Pb Pr"},
    "vga": {"name": "VGA"},
}

SOURCES_CURRENT: SourceCurrentType = {"id": "hdmi1"}

CHANNELS: ChannelsType = {
    "fingerprint-1": {
        "preset": "1",
        "name": "Flower",
    },
    "fingerprint-2": {"preset": "2", "name": "Moving Colourbar"},
    "fingerprint-3": {"preset": "12.3", "name": "Irdeto scrambled"},
    "fingerprint-4": {"preset": "4", "name": "Codec 16:9 scrambled"},
}

CHANNELS_CURRENT: ChannelsCurrentType = {"id": "fingerprint-1"}

CHANNELLISTS = {
    "fingerprint-1": {
        "name": "All TV channels",
        "source": "tv",
    },
    "fingerprint-2": {
        "name": "Favourite TV channels",
        "source": "tv",
    },
    "fingerprint-3": {
        "name": "Radio TV channels",
        "source": "tv",
    },
    "fingerprint-4": {
        "name": "Non-radio TV channels",
        "source": "tv",
    },
    "fingerprint-5": {
        "name": "All satellite channels",
        "source": "satellite",
    },
    "fingerprint-6": {
        "name": "Favourite satellite channels",
        "source": "satellite",
    },
    "fingerprint-7": {
        "name": "Radio satellite channels",
        "source": "satellite",
    },
    "fingerprint-8": {
        "name": "Non-radio satellite channels",
        "source": "satellite",
    },
}

VOLUME = {"muted": False, "current": 18, "min": 0, "max": 60}

AMBILIGHT = {
    "mode": {"current": "internal"},
    "topology": {"layers": 1, "left": 4, "top": 0, "right": 4, "bottom": 0},
    "measured": {
        "layer1": {
            "left": {
                "0": {"r": 56, "g": 43, "b": 40},
                "1": {"r": 94, "g": 81, "b": 77},
                "2": {"r": 76, "g": 70, "b": 60},
                "3": {"r": 43, "g": 37, "b": 26},
            },
            "top": {},
            "right": {
                "0": {"r": 69, "g": 70, "b": 58},
                "1": {"r": 124, "g": 120, "b": 100},
                "2": {"r": 83, "g": 87, "b": 90},
                "3": {"r": 50, "g": 49, "b": 51},
            },
            "bottom": {},
        }
    },
    "processed": {
        "layer1": {
            "left": {
                "0": {"r": 37, "g": 77, "b": 182},
                "1": {"r": 53, "g": 87, "b": 186},
                "2": {"r": 64, "g": 96, "b": 188},
                "3": {"r": 19, "g": 67, "b": 188},
            },
            "top": {},
            "right": {
                "0": {"r": 32, "g": 79, "b": 188},
                "1": {"r": 83, "g": 110, "b": 188},
                "2": {"r": 113, "g": 110, "b": 112},
                "3": {"r": 32, "g": 76, "b": 188},
            },
            "bottom": {},
        }
    },
    "cached": {
        "layer1": {
            "left": {
                "0": {"r": 0, "g": 0, "b": 0},
                "1": {"r": 0, "g": 0, "b": 0},
                "2": {"r": 0, "g": 0, "b": 0},
                "3": {"r": 0, "g": 0, "b": 0},
            },
            "top": {},
            "right": {
                "0": {"r": 0, "g": 0, "b": 0},
                "1": {"r": 0, "g": 0, "b": 0},
                "2": {"r": 0, "g": 0, "b": 0},
                "3": {"r": 0, "g": 0, "b": 0},
            },
            "bottom": {},
        }
    },
}

POWERSTATE_65PUS6121 = {"powerstate": "On"}

POWERSTATE = {"65PUS6121": POWERSTATE_65PUS6121}

CONTEXT_65PUS6121 = {
    "activity": "WatchExtension",
    "menu": "Source Menu",
    "Recording": "Off",
}

CONTEXT = {"65PUS6121": CONTEXT_65PUS6121}




from haphilipsjs import AUTH_SHARED_KEY, hmac_signature
import pytest

@pytest.mark.parametrize(
    ("timestamp", "auth_key", "pin", "auth_signature"),
    [
        (1128, "um27pghe7wyo0ysu", "0667", "rdPzC+bdWQHQMcwe3X+LddeG/tA=")
    ]
)
async def test_hmac(timestamp, auth_key, pin, auth_signature):
    
    assert hmac_signature(AUTH_SHARED_KEY, str(timestamp), pin) == auth_signature


import curses
import platform
import json
import sys
import pprint

from . import PhilipsTV
from haphilipsjs.typing import AmbilightCurrentConfiguration

import asyncio
from ast import literal_eval

pp = pprint.PrettyPrinter(indent=2)

async def monitor_run(stdscr: curses.window, tv: PhilipsTV):

    stdscr.clear()
    stdscr.timeout(1000)

    await tv.update()

    def get_application_name():

        if tv.applications:
            for app in tv.applications.values():
                if app.get("intent") == tv.application:
                    return app.get("label")

        if tv.application:
            return tv.application.get("component", {}).get("className")

    while True:

        stdscr.clear()
        stdscr.addstr(0, 0, "Source")
        if tv.source_id:
            stdscr.addstr(1, 0, await tv.getSourceName(tv.source_id) or "")

        stdscr.addstr(3, 15, "Menu Version")
        if tv.channel_id:
            stdscr.addstr(4, 15, str(tv.settings_version))

        stdscr.addstr(0, 15, "Channel")
        if tv.channel_id:
            stdscr.addstr(1, 15, await tv.getChannelName(tv.channel_id) or "")


        stdscr.addstr(0, 30, "Application")
        if tv.application:
            stdscr.addstr(1, 30, get_application_name() or "")

        stdscr.addstr(0, 45, "Context")
        if tv.context:
            stdscr.addstr(1, 45, tv.context.get("level1", ""))
            stdscr.addstr(2, 45, tv.context.get("level2", ""))
            stdscr.addstr(3, 45, tv.context.get("level3", ""))
            stdscr.addstr(4, 45, tv.context.get("data", ""))


        stdscr.addstr(0, 70, "Channels")
        for idx, channel  in enumerate(tv.channels_current):
            stdscr.addstr(1+idx, 70, str(channel.get("name", channel.get("ccid", ""))))


        def print_pixels(side, offset_y, offset_x):
            stdscr.addstr(offset_y, offset_x, "{}".format(side))
            stdscr.addstr(offset_y + 1, offset_x, "-----------")
            if side not in layer:
                return
            for pixel, data in layer[side].items():
                stdscr.addstr(
                    offset_y + 2 + int(pixel),
                    offset_x,
                    "{:>3} {:>3} {:>3}".format(data["r"], data["g"], data["b"]),
                )

        ambilight = await tv.getAmbilightProcessed()
        if ambilight:
            for idx, layer in enumerate(ambilight.values()):
                offset_y = 6 + 6 * idx
                print_pixels("left", offset_y, 0)
                print_pixels("top", offset_y, 15)
                print_pixels("right", offset_y, 30)
                print_pixels("bottom", offset_y, 45)

        stdscr.refresh()
        if not await tv.notifyChange():
            await tv.update()

        key = stdscr.getch()

        if key == ord("q"):
            break


async def monitor(tv):

    try:
        stdscr = curses.initscr()

        curses.noecho()
        curses.cbreak()

        stdscr.keypad(True)

        try:
            curses.start_color()
        except:
            pass

        return await monitor_run(stdscr, tv)
    finally:
        # Set everything back to normal
        if "stdscr" in locals():
            stdscr.keypad(False)  # type: ignore
            curses.echo()
            curses.nocbreak()
            curses.endwin()

    curses.wrapper(monitor_run, tv)


async def main():
    import argparse
    import logging
    import haphilipsjs
    import ast

    logger = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(prog="python -m haphilipsjs")
    parser.add_argument(
        "-d",
        "--debug",
        help="Debug output",
        dest="debug",
        action="store_true",
        default=False,
    )
    parser.add_argument("-i", "--host", dest="host", required=True)
    parser.add_argument("-a", "--api", dest="api", required=True)
    parser.add_argument(
        "-s",
        "--secured_transport",
        dest="secured_transport",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-u", "--username", dest="username", help="Username", required=False
    )
    parser.add_argument(
        "-p", "--password", dest="password", help="Password", required=False
    )

    subparsers = parser.add_subparsers(help="commands", dest="command")

    status = subparsers.add_parser("status", help="Show current tv status")

    mon = subparsers.add_parser("monitor", help="Monitor current tv status")

    ambilight = subparsers.add_parser("ambilight", help="Control ambilight")
    ambilight.add_argument(
        "--ambilight_mode", dest="ambilight_mode", type=str, required=False
    )
    ambilight.add_argument(
        "--ambilight_cached",
        dest="ambilight_cached",
        type=ast.literal_eval,
        required=False,
    )
    ambilight.add_argument(
        "--style", dest="ambilight_style", type=str, required=False
    )
    ambilight.add_argument(
        "--setting", dest="ambilight_setting", type=str, required=False
    )

    pair = subparsers.add_parser("pair", help="Pair with tv")

    get = subparsers.add_parser("get", help="Get data from endpoint")
    get.add_argument("path", help="Sub path to grab from tv")

    post = subparsers.add_parser("post", help="Post data to endpoint")
    post.add_argument("path", help="Sub path to grab from tv")
    post.add_argument("data", help="Json data to post")

    strings = subparsers.add_parser("strings", help="Get translations")
    strings.add_argument("string_ids", help="List of string id to request", nargs="+")
    strings.add_argument("--language", help="Language code to request")
    strings.add_argument("--country", help="Country code to request")
    strings.add_argument("--variant", help="Variant")

    settings = subparsers.add_parser("settings", help="Get setting")

    settings_commands = settings.add_subparsers(help="settings commands", dest="settings_command")
    settings_get = settings_commands.add_parser("get", help="Get current setting")
    settings_get.add_argument("node_id", help="node id to request", type=int)

    settings_structure = settings_commands.add_parser("structure", help="Get settings structure")

    settings_set = settings_commands.add_parser("set", help="Post a settings value")
    settings_set.add_argument("node_id", help="node id to request", type=int)
    settings_set.add_argument("data", help="Json data to post", type=json.loads)

    markdown = subparsers.add_parser("markdown", help="Print markdown for commandline")
    notify = subparsers.add_parser("notify", help="Look at notify changes, use with debug")


    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)
    haphilipsjs.TIMEOUT = 60.0
    tv = haphilipsjs.PhilipsTV(
        args.host,
        int(args.api),
        username=args.username,
        password=args.password,
        secured_transport=args.secured_transport,
    )
    try:
        await run(args, parser, tv)
    finally:
        await tv.aclose()


async def run(args, parser, tv: PhilipsTV):
    if args.command == "status":
        await tv.update()

        print("System: {}".format(tv.system))

        # Simulating homeassistant/components/media_player/philips_js.py
        print("Source: {}".format(await tv.getSourceName(tv.source_id)))
        if tv.sources:
            print(
                "Sources: {}".format(
                    ", ".join(
                        [
                            await tv.getSourceName(srcid) or "None"
                            for srcid in tv.sources
                        ]
                    )
                )
            )
        print(
            "Channel: {}: {}".format(
                tv.channel_id, await tv.getChannelName(tv.channel_id)
            )
        )
        if tv.channels:
            print(
                "Channels: {}".format(
                    ", ".join(
                        [
                            await tv.getChannelName(ccid) or "None"
                            for ccid in list(tv.channels.keys())
                        ]
                    )
                )
            )
        print("Context: {}".format(tv.context))

        print("Application:")
        pp.pprint(tv.application)

        print("Applications:")
        pp.pprint(list(tv.applications.values()))

        print("Power State: {}".format(tv.powerstate))
        print("Screen State: {}".format(tv.screenstate))

        await tv.getAmbilightPower()
        print("Ambilight power: {}".format(tv.ambilight_power))
        print("Ambilight mode: {}".format(tv.ambilight_mode))
        print("Ambilight topology:")
        pp.pprint(await tv.getAmbilightTopology())
        print("Ambilight processed:")
        pp.pprint(await tv.getAmbilightProcessed())
        print("Ambilight measured:")
        pp.pprint(await tv.getAmbilightMeasured())
        print("Ambilight styles:")
        pp.pprint(list(tv.ambilight_styles.values()))
        print("Ambilight currentconfiguration:")
        pp.pprint(tv.ambilight_current_configuration)
        print("Ambilight+Hue State: {}".format(tv.huelamp_power))

    elif args.command == "ambilight":
        await tv.getSystem()
        await tv.setTransport(tv.secured_transport, tv.api_version_detected)

        if args.ambilight_mode:
            if not await tv.setAmbilightMode(args.ambilight_mode):
                print("Failed to set mode")

        if args.ambilight_cached:
            if not await tv.setAmbilightCached(args.ambilight_cached):
                print("Failed to set ambilight cached")

        current: AmbilightCurrentConfiguration = {}
        if args.ambilight_style:
            current["styleName"] = args.ambilight_style
        if args.ambilight_setting:
            current["menuSetting"] = args.ambilight_setting
        if current:
            current["isExpert"] = False

            if await tv.setAmbilightCurrentConfiguration(current) is False:
                print("Failed to set ambilight config")


    elif args.command == "monitor":
        await monitor(tv)

    elif args.command == "pair":
        await tv.getSystem()
        await tv.setTransport(tv.secured_transport, tv.api_version_detected)
        state = await tv.pairRequest(
            "ha-philipsjs", "ha-philipsjs", platform.node(), platform.system(), "native"
        )

        pin = input("Please enter pin displayed on screen: ")

        res = await tv.pairGrant(state, pin)

        print(f"Username: {res[0]}")
        print(f"Password: {res[1]}")

    elif args.command == "get":
        res = await tv.getReq(args.path)
        json.dump(res, sys.stdout, indent=2, ensure_ascii=False)

    elif args.command == "post":
        res = await tv.postReq(args.path, literal_eval(args.data))
        json.dump(res, sys.stdout, indent=2, ensure_ascii=False)

    elif args.command == "strings":
        res = await tv.getStrings(args.string_ids, args.language, args.country, args.variant)
        json.dump(res, sys.stdout, indent=2, ensure_ascii=False)

    elif args.command == "settings":
        if args.settings_command == "structure":
            res = await tv.getMenuItemsSettingsStructure(force=True)
            json.dump(res, sys.stdout, indent=2, ensure_ascii=False)
        elif args.settings_command == "get":
            res = await tv.getMenuItemsSettingsCurrentValue([args.node_id], force=True)
            json.dump(res[args.node_id], sys.stdout, indent=2, ensure_ascii=False)
            print()
        elif args.settings_command == "set":
            res = await tv.postMenuItemsSettingsUpdateData({args.node_id: args.data}, force=True)
            json.dump(res, sys.stdout, indent=2, ensure_ascii=False)
            print()
    elif args.command == "markdown":
        try:
            import argmark  # type: ignore
        except ImportError:
            print("Unable to find argmmark")
            return

        argmark.md_help(parser)

    elif args.command == "notify":
        await tv.update()
        while await tv.notifyChange() is not None:
            pass

if __name__ == "__main__":
    asyncio.run(main())



import haphilipsjs
import pytest
import httpx
import respx
import json
from unittest.mock import patch
from typing import Dict, NamedTuple, cast

from haphilipsjs.data.v6 import (
    AMBILIGHT,
    ACTIVITIES_TV,
    AMBILIGHT_CURRENT_CONFIGURATION,
    AMBILIGHT_SUPPORTED_STYLES,
    AMBILIGHT_SUPPORTED_STYLES_EXTENDED,
    APPLICATIONS,
    CHANNELDB_TV_ANDROID,
    CHANNELDB_TV_SAPHI,
    CHANNELDB_TV_CHANNELLISTS_ALL,
    CHANNELDB_TV_FAVORITELISTS_ALLTER,
    CHANNELDB_TV_FAVORITELISTS_1,
    ACTIVITIES_CURRENT,
    CONTEXT,
    MENUITEMS_SETTINGS_STRUCTURE,
    MENUITEMS_SETTINGS_CURRENT,
    POWERSTATE,
    SYSTEM_ANDROID_DECRYPTED,
    SYSTEM_ANDROID_ENCRYPTED,
    SYSTEM_SAPHI_DECRYPTED,
    SYSTEM_SAPHI_ENCRYPTED,
    VOLUME,
    SCREENSTATE,
    HUELAMPPOWER,
    RECORDINGS_LIST,
    PAIR_REQUEST_DEVICE,
    PAIR_REQUEST_SAPHI,
    PAIR_REQUEST_ANDROID,
    PAIR_RESPONSE,
)
from haphilipsjs.typing import StringsRequest, Strings

MOCK_ANDROID_SOURCES = {
    "content://android.media.tv/channel": {"name": "Watch TV"},
    "content://android.media.tv/passthrough/com.mediatek.tvinput%2F.hdmi.HDMIInputService%2FHW5": {
        "name": "HDMI 1"
    },
    "content://android.media.tv/passthrough/com.mediatek.tvinput%2F.hdmi.HDMIInputService%2FHW6": {
        "name": "HDMI 2"
    },
    "content://android.media.tv/passthrough/com.mediatek.tvinput%2F.hdmi.HDMIInputService%2FHW7": {
        "name": "HDMI 3"
    },
    "content://android.media.tv/passthrough/com.mediatek.tvinput%2F.hdmi.HDMIInputService%2FHW8": {
        "name": "HDMI 4"
    },
}

MOCK_SAPHI_SOURCES = {}


class Param(NamedTuple):
    type: str
    base: str


@pytest.fixture(params=["android", "saphi"], name="param")
async def param_fixture(request):
    if request.param == "android":
        return Param(request.param, "https://127.0.0.1:1926/6")
    elif request.param == "saphi":
        return Param(request.param, "http://127.0.0.1:1925/6")
    else:
        raise Exception


@pytest.fixture
async def client_mock(param: Param):
    with respx.mock:
        if param.type == "android":
            respx.get(f"{param.base}/system").respond(
                json=cast(Dict, SYSTEM_ANDROID_ENCRYPTED)
            )
            respx.get("http://127.0.0.1:1925/6/system").respond(
                json={}
            )
            respx.get(f"{param.base}/channeldb/tv").respond(
                json=cast(Dict, CHANNELDB_TV_ANDROID)
            )
        elif param.type == "saphi":
            respx.get(f"{param.base}/system").respond(
                json=cast(Dict, SYSTEM_SAPHI_ENCRYPTED)
            )
            respx.get(f"{param.base}/channeldb/tv").respond(
                json=cast(Dict, CHANNELDB_TV_SAPHI)
            )
        else:
            raise Exception

        respx.get(f"{param.base}/channeldb/tv/channelLists/all").respond(
            json=cast(Dict, CHANNELDB_TV_CHANNELLISTS_ALL)
        )

        respx.get(f"{param.base}/channeldb/tv/favoriteLists/allter").respond(
            json=cast(Dict, CHANNELDB_TV_FAVORITELISTS_ALLTER)
        )

        respx.get(f"{param.base}/channeldb/tv/favoriteLists/1").respond(
            json=cast(Dict, CHANNELDB_TV_FAVORITELISTS_1)
        )

        respx.get(f"{param.base}/activities/current").respond(
            json=cast(Dict, ACTIVITIES_CURRENT)
        )
        respx.get(f"{param.base}/activities/tv").respond(json=cast(Dict, ACTIVITIES_TV[param.type]))
        respx.get(f"{param.base}/applications").respond(json=cast(Dict, APPLICATIONS))
        respx.get(f"{param.base}/powerstate").respond(json=POWERSTATE)
        respx.get(f"{param.base}/screenstate").respond(json=SCREENSTATE)
        respx.get(f"{param.base}/context").respond(json=cast(Dict, CONTEXT))
        respx.get(f"{param.base}/audio/volume").respond(json=VOLUME)
        respx.get(f"{param.base}/ambilight/mode").respond(json=AMBILIGHT["mode"])
        respx.get(f"{param.base}/ambilight/topology").respond(
            json=AMBILIGHT["topology"]
        )
        respx.get(f"{param.base}/ambilight/measured").respond(
            json=AMBILIGHT["measured"]
        )
        respx.get(f"{param.base}/ambilight/processed").respond(
            json=AMBILIGHT["processed"]
        )
        respx.get(f"{param.base}/ambilight/cached").respond(json=AMBILIGHT["cached"])
        respx.get(f"{param.base}/ambilight/power").respond(json={"power": "On"})
        respx.get(f"{param.base}/ambilight/supportedstyles").respond(
            json=cast(Dict, AMBILIGHT_SUPPORTED_STYLES)
        )
        respx.get(f"{param.base}/ambilight/currentconfiguration").respond(
            json=cast(Dict, AMBILIGHT_CURRENT_CONFIGURATION)
        )
        respx.get(f"{param.base}/HueLamp/power").respond(json=HUELAMPPOWER)
        respx.get(f"{param.base}/menuitems/settings/structure").respond(
            json=cast(Dict, MENUITEMS_SETTINGS_STRUCTURE)
        )
        respx.post(f"{param.base}/menuitems/settings/current").respond(
            json=cast(Dict, MENUITEMS_SETTINGS_CURRENT)
        )
        if param.type == "android":
            client = haphilipsjs.PhilipsTV(
                "127.0.0.1", api_version=6, secured_transport=True
            )
        elif param.type == "saphi":
            client = haphilipsjs.PhilipsTV(
                "127.0.0.1", api_version=6, secured_transport=False
            )
        else:
            raise Exception
        respx.get(f"{param.base}/recordings/list").respond(
            json=RECORDINGS_LIST
        )

        yield client
        await client.aclose()


async def test_basic_data(client_mock, param: Param):
    """Test for basic data"""
    await client_mock.update()
    assert client_mock.on == True

    if param.type == "android":
        assert client_mock.system == SYSTEM_ANDROID_DECRYPTED
        assert client_mock.sources == MOCK_ANDROID_SOURCES
        assert client_mock.applications == {
            app["id"]: app for app in APPLICATIONS["applications"] if "id" in app
        }
        assert (
            client_mock.application_id
            == "org.droidtv.nettv.market.MarketMainActivity-org.droidtv.nettv.market"
        )
        assert client_mock.quirk_ambilight_mode_ignored == True
        assert client_mock.os_type == "MSAF_2019_P"

        assert client_mock.channel_list_id == "1"
        assert client_mock.channels_current == [
            {"ccid": 1649, "preset": "1"}
        ]

    elif param.type == "saphi":
        assert client_mock.system == SYSTEM_SAPHI_DECRYPTED
        assert client_mock.sources == MOCK_SAPHI_SOURCES
        assert client_mock.applications == {}
        assert client_mock.application_id == None
        assert client_mock.quirk_ambilight_mode_ignored == True
        assert client_mock.os_type == "Linux"

        assert client_mock.channel_list_id == "all"
        assert client_mock.channels_current == list(client_mock.channels.values())


    assert client_mock.channels == {
        "1648": {"ccid": 1648, "preset": "1", "name": "Irdeto scrambled"},
        "1649": {"ccid": 1649, "preset": "2"},
    }
    assert client_mock.powerstate == POWERSTATE["powerstate"]
    assert client_mock.screenstate == SCREENSTATE["screenstate"]
    assert client_mock.huelamp_power == HUELAMPPOWER["power"]


async def test_current_channel_none(client_mock, param):
    await client_mock.update()
    assert client_mock.channel_id == "1648"

    respx.get(f"{param.base}/activities/tv").respond(json={})
    await client_mock.update()
    assert client_mock.channel_id == None


async def test_get_channel_name(client_mock):
    """Verify that we can translate channel id to name"""
    await client_mock.update()
    assert await client_mock.getChannelName("1648") == "Irdeto scrambled"
    assert await client_mock.getChannelName("balha") == None


async def test_set_source(client_mock, param):
    """Verify that we can translate channel id to name"""
    await client_mock.update()

    if param.type == "android":
        route = respx.post(f"{param.base}/activities/launch").respond(json={})
        assert await client_mock.setSource(
            "content://android.media.tv/passthrough/com.mediatek.tvinput%2F.hdmi.HDMIInputService%2FHW5"
        )
        assert json.loads(route.calls[0].request.content) == {
            "intent": {
                "extras": {
                    "uri": "content://android.media.tv/passthrough/com.mediatek.tvinput%2F.hdmi.HDMIInputService%2FHW5"
                },
                "action": "org.droidtv.playtv.SELECTURI",
                "component": {
                    "packageName": "org.droidtv.playtv",
                    "className": "org.droidtv.playtv.PlayTvActivity",
                },
            }
        }
    elif param.type == "saphi":
        assert await client_mock.setSource("a") == False


async def test_timeout(client_mock, param):
    """Test that connect timeouts trigger tv to be considered off"""
    await client_mock.update()
    assert client_mock.on == True
    respx.get(f"{param.base}/audio/volume").mock(side_effect=httpx.ConnectTimeout)

    await client_mock.update()
    assert client_mock.on == False

    respx.get(f"{param.base}/audio/volume").respond(json=VOLUME)

    await client_mock.update()
    assert client_mock.on == True


async def test_volume(client_mock, param: Param):
    respx.get(f"{param.base}/audio/volume").respond(
        json={"muted": True, "current": 30, "min": 0, "max": 60}
    )

    await client_mock.update()
    assert client_mock.volume == 0.5
    assert client_mock.muted == True

    respx.get(f"{param.base}/audio/volume").respond(
        json={"muted": False, "current": 60, "min": 0, "max": 60}
    )

    await client_mock.update()
    assert client_mock.volume == 1.0
    assert client_mock.muted == False

    respx.get(f"{param.base}/audio/volume").respond(
        json={"muted": False, "current": 0, "min": 0, "max": 60}
    )

    await client_mock.update()
    assert client_mock.volume == 0.0
    assert client_mock.muted == False


async def test_set_volume(client_mock, param: Param):
    await client_mock.update()
    respx.post(f"{param.base}/audio/volume").respond(json={})

    await client_mock.setVolume(0.5)

    assert respx.calls[-1].request.url == f"{param.base}/audio/volume"
    assert json.loads(respx.calls[-1].request.content) == {
        "muted": False,
        "current": 30,
    }

    await client_mock.setVolume(1.0, True)

    assert json.loads(respx.calls[-1].request.content) == {
        "muted": True,
        "current": 60,
    }

    await client_mock.setVolume(0.0, True)

    assert json.loads(respx.calls[-1].request.content) == {
        "muted": True,
        "current": 0,
    }

    await client_mock.setVolume(None, True)

    assert json.loads(respx.calls[-1].request.content) == {
        "muted": True,
    }


async def test_set_channel(client_mock, param: Param):
    respx.post(f"{param.base}/activities/tv").respond(json={})

    await client_mock.setChannel(1649)

    assert respx.calls[-1].request.url == f"{param.base}/activities/tv"
    assert json.loads(respx.calls[-1].request.content) == {
        "channelList": {"id": "all"},
        "channel": {"ccid": 1649},
    }


async def test_send_key(client_mock, param: Param):
    respx.post(f"{param.base}/input/key").respond(json={})

    await client_mock.update()
    await client_mock.sendKey("Standby")

    assert respx.calls[-1].request.url == f"{param.base}/input/key"
    assert json.loads(respx.calls[-1].request.content) == {
        "key": "Standby",
    }

    assert client_mock.audio_volume["muted"] == False
    await client_mock.sendKey("Mute")
    assert client_mock.audio_volume["muted"] == True

    assert client_mock.audio_volume["current"] == 18
    await client_mock.sendKey("VolumeUp")
    assert client_mock.audio_volume["current"] == 19
    await client_mock.sendKey("VolumeDown")
    assert client_mock.audio_volume["current"] == 18


async def test_send_key_off(client_mock, param: Param):
    respx.post(f"{param.base}/input/key").mock(side_effect=httpx.ConnectTimeout)

    with pytest.raises(haphilipsjs.ConnectionFailure):
        await client_mock.sendKey("Standby")


async def test_ambilight_mode(client_mock, param):
    await client_mock.getSystem()

    respx.post(f"{param.base}/ambilight/mode").respond(json={})
    await client_mock.setAmbilightMode("internal")

    if client_mock.quirk_ambilight_mode_ignored:
        assert respx.calls[-2].request.url == f"{param.base}/ambilight/mode"
        assert json.loads(respx.calls[-2].request.content) == {
            "current": "internal",
        }
        assert respx.calls[-1].request.url == f"{param.base}/ambilight/mode"
        assert json.loads(respx.calls[-1].request.content) == {
            "current": "internal_invalid",
        }
    else:
        assert respx.calls[-1].request.url == f"{param.base}/ambilight/mode"
        assert json.loads(respx.calls[-1].request.content) == {
            "current": "internal",
        }

    assert await client_mock.getAmbilightMode()
    assert client_mock.ambilight_mode == "internal"

    assert await client_mock.setAmbilightMode("manual")
    assert client_mock.ambilight_mode == "manual"

    if client_mock.quirk_ambilight_mode_ignored:
        assert await client_mock.getAmbilightMode()
        assert client_mock.ambilight_mode == "manual"

    assert await client_mock.setAmbilightMode("interal")
    assert await client_mock.getAmbilightMode()


async def test_ambilight_power(client_mock, param):
    route = respx.post(f"{param.base}/ambilight/power").respond(json={})

    assert client_mock.ambilight_power == None
    await client_mock.update()
    await client_mock.getAmbilightPower()

    assert client_mock.ambilight_power == "On"

    await client_mock.setAmbilightPower("On")
    assert json.loads(route.calls[0].request.content) == {"power": "On"}

    await client_mock.setAmbilightPower("Off")
    assert json.loads(route.calls[1].request.content) == {"power": "Off"}


async def test_ambilight_topology(client_mock):
    assert await client_mock.getAmbilightTopology() == AMBILIGHT["topology"]


async def test_ambilight_measured(client_mock):
    assert await client_mock.getAmbilightMeasured() == AMBILIGHT["measured"]


async def test_ambilight_processed(client_mock):
    assert await client_mock.getAmbilightProcessed() == AMBILIGHT["processed"]


async def test_ambilight_cached(client_mock, param: Param):
    assert await client_mock.getAmbilightCached() == AMBILIGHT["cached"]

    respx.post(f"{param.base}/ambilight/cached").respond(json={})

    data = {"r": 100, "g": 210, "b": 30}

    await client_mock.setAmbilightCached(data)

    assert respx.calls[-2].request.url == f"{param.base}/ambilight/cached"

    assert respx.calls[-2].request.url == f"{param.base}/ambilight/cached"
    assert json.loads(respx.calls[-2].request.content) == data


async def test_ambilight_modes(client_mock, param):
    await client_mock.getSystem()
    if param.type == "android":
        assert client_mock.ambilight_modes == ["internal", "manual", "expert", "lounge"]
    elif param.type == "saphi":
        assert client_mock.ambilight_modes == ["internal", "manual", "expert"]


async def test_ambilight_current_configuration(client_mock, param):
    respx.post(f"{param.base}/ambilight/currentconfiguration").respond(json={})

    await client_mock.getSystem()
    await client_mock.getAmbilightCurrentConfiguration()

    data = {
        "isExpert": False,
        "menuSetting": "STANDARD",
        "stringValue": "Standard",
        "styleName": "FOLLOW_VIDEO",
    }

    assert client_mock.ambilight_current_configuration == data

    await client_mock.setAmbilightCurrentConfiguration(data)
    assert json.loads(respx.calls[-1].request.content) == data


async def test_ambilight_supported_stypes(client_mock, param):
    await client_mock.getSystem()
    await client_mock.getAmbilightSupportedStyles()
    assert client_mock.ambilight_styles == AMBILIGHT_SUPPORTED_STYLES_EXTENDED[param.type]


async def test_menu_items_current(client_mock: haphilipsjs.PhilipsTV, param):
    await client_mock.getSystem()

    MOCK_VALUE_1 = {
        "Nodeid": 2131230778,
        "Controllable": False,
        "Available": True,
        "string_id": "Inschakelen",
        "data": {
            "value": True
        }
    }
    MOCK_VALUE_2 = {
        "Nodeid": 2131230779,
        "Controllable": False,
        "Available": True,
        "string_id": "Inschakelen 2",
        "data": {
            "value": True
        }
    }

    route = respx.post(f"{param.base}/menuitems/settings/current")
    route.side_effect = [
        httpx.Response(
            status_code=200,
            json={
                "values": [{"value": MOCK_VALUE_1}],
                "version": 0
            }
        ),
        httpx.Response(
            status_code=200,
            json={
                "values": [{"value": MOCK_VALUE_2}],
                "version": 0
            }
        ),
    ]
 
    with patch.object(haphilipsjs, "MAXIMUM_ITEMS_IN_REQUEST", new=1):
        data = await client_mock.getMenuItemsSettingsCurrentValue([2131230778, 2131230779])
        if param.type == "android":
            assert data == {
                2131230778: MOCK_VALUE_1,
                2131230779: MOCK_VALUE_2,
            }
        elif param.type == "saphi":
            assert data == {
                2131230778: None,
                2131230779: None
            }

async def test_get_strings_cached(client_mock: haphilipsjs.PhilipsTV, param: Param):
    await client_mock.getSystem()
    assert client_mock.strings == {}

    with respx.mock as mock:
        def mock_strings(request: StringsRequest, response: Strings):
            return mock.post(f"{param.base}/strings", json=request).respond(json=cast(Dict, response))

        mock_strings({
                "strings": [{"string_id": "1"}, {"string_id": "2"}],
                "locale": {
                    "language": "",
                    "country": "",
                    "variant": "",
                }
            },
            {
                "translations": [
                    { "string_id": "1", "string_translation": "ONE"},
                    { "string_id": "2", "string_translation": "TWO"}
                ]
            }
        )

        res = await client_mock.getStringsCached(["1","2"])
        assert res == {"1": "ONE", "2": "TWO"}
        assert client_mock.strings == {"1": "ONE", "2": "TWO"}

        mock_strings({
                "strings": [{"string_id": "3"}, {"string_id": "4"}],
                "locale": {
                    "language": "",
                    "country": "",
                    "variant": "",
                },
            },
            {
                "translations": [
                    { "string_id": "3", "string_translation": "THREE"},
                    { "string_id": "4", "string_translation": "FOUR"}
                ]
            }
        )

        res = await client_mock.getStringsCached(["1","3","4"])
        assert res == {"1": "ONE", "2": "TWO", "3": "THREE", "4": "FOUR"}
        assert client_mock.strings == {"1": "ONE", "2": "TWO", "3": "THREE", "4": "FOUR"}


async def test_buggy_json():
    assert haphilipsjs.decode_xtv_json("") == {}
    assert haphilipsjs.decode_xtv_json("}") == {}
    assert haphilipsjs.decode_xtv_json("Context Service not started") == {}
    assert haphilipsjs.decode_xtv_json('{,"a":{}}') == {"a": {}}
    assert haphilipsjs.decode_xtv_json('{"a":{},}') == {"a": {}}
    assert haphilipsjs.decode_xtv_json('{"a":{},,,"b":{}}') == {"a": {}, "b": {}}
    with pytest.raises(haphilipsjs.NoneJsonData):
        haphilipsjs.decode_xtv_json("Plain text data")

async def test_get_recordings(client_mock):
    """Verify that we can read back selected recording values"""
    await client_mock.update()

    recording_ongoing = False
    recording_new = 0

    assert client_mock.recordings_list["version"] == "253.91"

    for rec in client_mock.recordings_list["recordings"]:
        if rec["RecordingId"] == 36:
            rec_time_planned = rec["StartTime"]
            rec_margin_start = rec["MarginStart"]
            rec_time = rec_time_planned - rec_margin_start
            assert rec_time == 1676833531
            assert rec["RecordingType"] == "RECORDING_ONGOING"
            assert rec["RecError"] == "REC_ERROR_NONE"
        
        if rec["RecordingType"] == "RECORDING_ONGOING":
                recording_ongoing = True

        if rec["RecordingType"] == "RECORDING_NEW":
            recording_new += 1

    assert recording_ongoing == True
    assert recording_new == 1

async def test_pair_request(client_mock: haphilipsjs.PhilipsTV, param: Param):
    await client_mock.update()

    if param.type == "android":
        post_json = PAIR_REQUEST_ANDROID
    elif param.type == "saphi":
        post_json = PAIR_REQUEST_SAPHI
    else:
        pytest.skip("Unknown pair data")

    device = post_json["device"]

    with respx.mock as mock:
        route = mock.post(f"{param.base}/pair/request")
        route.respond(
            json=cast(Dict, PAIR_RESPONSE)
        )
        data = await client_mock.pairRequest(
            app_id=device["app_id"],
            app_name=device["app_name"],
            device_name=device["device_name"],
            device_os=device["device_os"],
            type=device["type"],
            device_id = device["id"]
        )
        
        assert json.loads(route.calls.last.request.content) == post_json
        assert data == {
            "device": device,
            "auth_key": PAIR_RESPONSE["auth_key"],
            "timestamp": PAIR_RESPONSE["timestamp"],
        }



from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union, cast, overload, Iterable, TypeVar
import httpx
import logging
import json
import itertools
from urllib.parse import quote
from secrets import token_bytes, token_hex
from base64 import b64decode, b64encode

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.padding import PKCS7
from cryptography.hazmat.primitives.hashes import SHA1
from cryptography.hazmat.primitives.hmac import HMAC

from .typing import (
    ActivitesTVType,
    AmbilightCurrentConfiguration,
    AmbilightLayersType,
    AmbilightSupportedStyleType,
    AmbilightSupportedStylesType,
    ApplicationIntentType,
    ApplicationsType,
    ChannelDbTv,
    ChannelListType,
    ChannelsCurrentType,
    ChannelsType,
    ChannelType,
    ContextType,
    FavoriteListType,
    RecordingsListed,
    MenuItemsSettingsCurrent,
    MenuItemsSettingsCurrentPost,
    MenuItemsSettingsCurrentValueValue,
    MenuItemsSettingsEntry,
    MenuItemsSettingsNode,
    MenuItemsSettingsStructure,
    MenuItemsSettingsUpdate,
    MenuItemsSettingsValueData,
    Strings,
    StringsRequest,
    SystemType,
    ApplicationType,
)
from .auth import CachedDigestAuth

LOG = logging.getLogger(__name__)
TIMEOUT = 20.0
TIMEOUT_CONNECT = 5.0
TIMEOUT_NOTIFYREAD = 130
DEFAULT_API_VERSION = 1

AUTH_SHARED_KEY = b64decode(
    "ZmVay1EQVFOaZhwQ4Kv81ypLAZNczV9sG4KkseXWn1NEk6cXmPKO/MCa9sryslvLCFMnNe4Z4CPXzToowvhHvA=="
)
"""Key used for hmac signatures and decoding of cbc data."""

TV_PLAYBACK_INTENTS = [
    {
        "component": {
            "className": "org.droidtv.playtv.PlayTvActivity",
            "packageName": "org.droidtv.playtv",
        }
    }
]

HTTP_PORT = 1925
HTTPS_PORT = 1926
MAXIMUM_ITEMS_IN_REQUEST = 50
IGNORED_JSON_RESPONSES = {"", "Context Service not started", "}"}

def hmac_signature(key: bytes, timestamp: str, data: str):
    """Calculate a timestamped signature."""
    hmac = HMAC(key, SHA1())
    hmac.update(timestamp.encode("utf-8"))
    hmac.update(data.encode("utf-8"))
    return b64encode(hmac.finalize()).decode("utf-8")


def cbc_decode(key: bytes, data: str):
    """Decoded encrypted fields based on shared key."""
    if data == "":
        return ""
    raw = b64decode(data)
    assert len(raw) >= 16, f"Lenght of data too short: '{data}'"
    decryptor = Cipher(algorithms.AES(key[0:16]), modes.CBC(raw[0:16])).decryptor()
    unpadder = PKCS7(128).unpadder()
    result = decryptor.update(raw[16:]) + decryptor.finalize()
    result = unpadder.update(result) + unpadder.finalize()
    return result.decode("utf-8")


def cbc_encode(key: bytes, data: str):
    """Decoded encrypted fields based on shared key."""
    raw = data.encode("utf-8")
    iv = token_bytes(16)
    encryptor = Cipher(algorithms.AES(key[0:16]), modes.CBC(iv)).encryptor()
    padder = PKCS7(128).padder()
    result = padder.update(raw) + padder.finalize()
    result = encryptor.update(result) + encryptor.finalize()
    return b64encode(iv + result).decode("utf-8")


def decode_xtv_json(text: str):
    if text in IGNORED_JSON_RESPONSES:
        LOG.debug("Ignoring invalid json %s", text)
        return {}

    try:
        data = json.loads(text)
    except json.decoder.JSONDecodeError:
        LOG.debug("Invalid json received, trying adjusted version")
        text = text.replace("{,", "{")
        text = text.replace(",}", "}")
        while (p := text.find(",,")) >= 0:
            text = text[:p] + text[p + 1 :]

        try:
            data = json.loads(text)
        except json.decoder.JSONDecodeError as exception:
            raise NoneJsonData(text) from exception

    return data


def decode_xtv_response(response: httpx.Response):
    try:
        text = response.text
    except UnicodeDecodeError:
        LOG.warning("Unable to decode unicode on endpoint, ignoring", exc_info=True)
        return {}

    if not response.headers.get("content-type", "").startswith("application/json"):
        if text:
            LOG.warning("Non json data: %s", text)
        return {}

    return decode_xtv_json(text)


_T = TypeVar("_T")

def chunked_iterator(size: int, iterable: Iterable[_T]) -> Iterable[Iterable[_T]]:
    """Create a chain of iterators that each will give a chunk of the original."""
    step = iter(iterable)
    while True:
        chunk_it = itertools.islice(step, size)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


PASSTHROUGH_URI = "content://android.media.tv/passthrough"


def passthrough_uri(data):
    return f"{PASSTHROUGH_URI}/{quote(data, safe='')}"


CHANNEL_URI = "content://android.media.tv/channel"


def channel_uri(channel):
    uri = CHANNEL_URI
    if channel is not None:
        uri += f"/{channel}"
    return uri


class GeneralFailure(Exception):
    """Base class for component failures."""


class ConnectionFailure(GeneralFailure):
    """Failed to connect to tv it's likely turned off."""


class ProtocolFailure(GeneralFailure):
    """Wrapper to contain erros that are the server closing a connection before response."""


class AutenticationFailure(GeneralFailure):
    """Wrapper to contain failures due to authentication."""


class PairingFailure(GeneralFailure):
    def __init__(self, data):
        super().__init__(f"Failed to start pairing: {data}")
        self.data = data


class NoneJsonData(GeneralFailure):
    """API Returned non json data when json was expected."""
    def __init__(self, data):
        super().__init__(f"Non json data received: {data}")
        self.data = data


T = TypeVar("T")


class PhilipsTV(object):

    channels: ChannelsType
    """All available channels, with ccid as key."""

    def __init__(
        self,
        host=None,
        api_version=DEFAULT_API_VERSION,
        secured_transport=None,
        username=None,
        password=None,
        verify=False,
        auth_shared_key=None,
        system = None,
        limits = None,
    ):
        self._host = host
        self._connfail = 0
        self.api_version = int(api_version)
        self.on = False
        self.name: Optional[str] = None
        self.system: Optional[SystemType] = system
        self.strings: dict[str, str] = {}
        self.sources = {}
        self.source_id = None
        self.audio_volume = None
        self.channels = {}
        self.channel: Optional[Union[ActivitesTVType, ChannelsCurrentType]] = None
        self.channel_lists: Dict[str, ChannelListType] = {}
        self.favorite_lists: Dict[str, FavoriteListType] = {}
        self.applications: Dict[str, ApplicationType] = {}
        self.application: Optional[ApplicationIntentType] = None
        self.context: Optional[ContextType] = None
        self.screenstate: Optional[str] = None
        self.settings: Optional[MenuItemsSettingsStructure] = None
        self.settings_version = 0
        self.ambilight_topology = None
        self.ambilight_mode_set = None
        self.ambilight_mode_raw: Optional[str] = None
        self.ambilight_cached: Optional[AmbilightLayersType] = None
        self.ambilight_measured: Optional[AmbilightLayersType] = None
        self.ambilight_processed: Optional[AmbilightLayersType] = None
        self.ambilight_power_raw: Optional[Dict] = None
        self.ambilight_styles: Dict[str, AmbilightSupportedStyleType] = {}
        self.ambilight_current_configuration: Optional[
            AmbilightCurrentConfiguration
        ] = None
        self.recordings_list: Optional[RecordingsListed] = None
        self.huelamp_power: Optional[str] = None
        self.powerstate = None
        if auth_shared_key:
            self.auth_shared_key = auth_shared_key
        else:
            self.auth_shared_key = AUTH_SHARED_KEY

        if secured_transport or self.secured_transport:
            self.protocol = "https"
        else:
            self.protocol = "http"

        timeout = httpx.Timeout(timeout=TIMEOUT, connect=TIMEOUT_CONNECT)
        if limits is None:
            limits = httpx.Limits(max_keepalive_connections=3, max_connections=3)
        self.session = httpx.AsyncClient(limits=limits, timeout=timeout, verify=False)
        self.session.headers["Accept"] = "application/json"

        if username and password:
            self.session.auth = CachedDigestAuth(username, password)

    @property
    def quirk_playpause_spacebar(self):
        """Does this tv need workaround for playpause."""
        if self.system:
            return self.system.get("os_type", "").startswith("MSAF_")
        else:
            return False

    @property
    def quirk_ambilight_styles_menuitems(self):
        if self.system:
            return self.system.get("os_type", "").startswith("MSAF_")
        else:
            return False

    @property
    def os_type(self):
        if self.system is None:
            return None

        # android system have in direclty in root
        os_type = self.system.get("os_type")
        if os_type:
            return os_type

        # saphi stores in in features
        os_type = (
            self.system.get("featuring", {}).get("systemfeatures", {}).get("os_type")
        )
        if os_type:
            return os_type

        return None

    @property
    def quirk_ambilight_mode_ignored(self):
        """Return if this tv need workaround for ambilight bugs.

        XTV app have bugs with their mode management for ambilight. It will
        forgot to actuate the command to set mode back to internal. But will actually
        do that if you give it an invalid mode.

        It will also not report a correct ambilight mode after being
        changed by call. So we need to remember last set mode.

        Versions known affected:
            - Android - 9.0.0
            - Saphi - 4.6.0.2

        Version known good
            - Legacy - QF1EU-0.150.102.0
        """

        os_type = self.os_type
        if os_type:
            if os_type.startswith("MSAF_"):
                return True

            if os_type == "Linux":
                return True

        return False

    @property
    def pairing_type(self):
        if self.system:
            return (
                self.system.get("featuring", {})
                .get("systemfeatures", {})
                .get("pairing_type")
            )
        else:
            return None

    @property
    def secured_transport(self) -> Optional[bool]:
        if self.system:
            return (
                self.system.get("featuring", {})
                .get("systemfeatures", {})
                .get("secured_transport")
                in ("true", True)
            )
        else:
            return None

    @property
    def notify_change_supported(self) -> Optional[str]:
        if self.system:
            return self.system.get("notifyChange", None)
        else:
            return None

    @overload
    def json_feature_supported(self, type: str) -> Optional[List[str]]:
        ...

    @overload
    def json_feature_supported(self, type: str, value: str) -> Optional[bool]:
        ...

    def json_feature_supported(self, type: str, value: Optional[str] = None):
        if self.system:
            features = cast(
                List[str],
                self.system.get("featuring", {}).get("jsonfeatures", {}).get(type, []),
            )
            if value:
                return value in features
            else:
                return features
        else:
            return None

    @property
    def api_version_detected(self) -> Optional[int]:
        if self.system:
            return cast(Optional[int], self.system.get("api_version", {}).get("Major"))
        else:
            return None

    @property
    def channel_active(self):
        if self.context and "level1" in self.context:
            return self.context["level1"] in ("WatchTv", "WatchSatellite")
        if self.context and "activity" in self.context:
            return self.context["activity"] in ("WatchTv", "WatchSatellite")
        if self.application:
            return self.application in TV_PLAYBACK_INTENTS
        if self.source_id in ("tv", "11", None):
            return self.channel_id is not None
        return False

    @property
    def application_id(self):
        if self.application and "component" in self.application:
            component = self.application["component"]
            app_id = f"{component.get('className', 'None')}-{component.get('packageName', 'None')}"
            if app_id in self.applications:
                return app_id
            else:
                return None
        else:
            return None

    @property
    def min_volume(self):
        if self.audio_volume:
            return int(self.audio_volume["min"])
        else:
            return None

    @property
    def max_volume(self):
        if self.audio_volume:
            return int(self.audio_volume["max"])
        else:
            return None

    @property
    def volume(self):
        if self.audio_volume and int(self.audio_volume["max"]):
            return self.audio_volume["current"] / int(self.audio_volume["max"])
        else:
            return None

    @property
    def muted(self):
        if self.audio_volume:
            return self.audio_volume["muted"]
        else:
            return None

    @property
    def channel_id(self):
        if self.api_version >= 5:
            r = cast(Optional[ActivitesTVType], self.channel)
            if not r:
                return None

            ccid = r.get("channel", {}).get("ccid")
            # it could be empty if HDMI is set
            if not ccid:
                return None
            return str(ccid)
        else:
            r = cast(Optional[ChannelsCurrentType], self.channel)
            if not r:
                return None

            if not self.channels.get(r["id"]):
                pos = r["id"].find("_")
                if pos > 0:
                    return r["id"][pos + 1 :]
            return r["id"]

    @property
    def channel_list_id(self) -> str:
        if self.api_version >= 5:
            if not self.channel:
                return "all"
            r = cast(ActivitesTVType, self.channel)
            channel_list = r.get("channelList")
            if not channel_list:
                return "all"
            return channel_list.get("id", "all")

        return "all"

    @property
    def channels_current(self) -> List[ChannelType]:
        """All channels in the current favorite list."""
        if self.api_version >= 5:
            favorite_list = self.favorite_lists.get(self.channel_list_id)
            if not favorite_list:
                return list(self.channels.values())

            return [
                {
                    **channel,
                    "preset": favorite.get("preset", "")
                }
                for favorite in favorite_list.get("channels", [])
                if (channel := self.channels.get(str(favorite.get("ccid"))))
            ]
        else:
            return [
                {
                    **channel,
                    "ccid": key
                }
                for key, channel in self.channels.items()
            ]

    @property
    def ambilight_modes(self):
        modes = ["internal", "manual", "expert"]

        if self.api_version >= 6:
            features = self.json_feature_supported("ambilight")
            if features:
                if "LoungeLight" in features:
                    modes.append("lounge")
                return modes
            else:
                return []
        else:
            return modes

    @property
    def ambilight_mode(self) -> Optional[str]:
        if self.quirk_ambilight_mode_ignored:
            if self.ambilight_mode_set:
                return self.ambilight_mode_set
        return self.ambilight_mode_raw

    @property
    def ambilight_power(self) -> Optional[str]:

        if self.api_version >= 5:
            if self.ambilight_power_raw is None:
                return None
            return self.ambilight_power_raw.get("power")

        if self.api_version == 1:
            if self.ambilight_cached is None:
                return None

            if self.ambilight_mode == "manual":
                if any(
                    color != 0
                    for layer in self.ambilight_cached.values()
                    for side in layer.values()
                    for pixel in side.values()
                    for color in pixel.values()
                ):
                    return "On"
                else:
                    return "Off"

            else:
                return "On"

        return None

    async def aclose(self) -> None:
        await self.session.aclose()

    def _url(self, path, protocol = None):
        if protocol is None:
            protocol = self.protocol

        if protocol == "https":
            port = HTTPS_PORT
        else:
            port = HTTP_PORT

        return f"{protocol}://{self._host}:{port}/{self.api_version}/{path}"

    async def getReq(self, path, protocol = None) -> Optional[Dict]:
        try:
            resp = await self.session.get(self._url(path, protocol = protocol))
            if resp.status_code == 401:
                raise AutenticationFailure("Authenticaion failed to device")

            if resp.status_code != 200:
                LOG.debug("Get failed: %s -> %d %s", path, resp.status_code, resp.text)
                return None

            LOG.debug("Get succeded: %s -> %s", path, resp.text)
            return decode_xtv_response(resp)
        except (httpx.ConnectTimeout, httpx.ConnectError) as err:
            raise ConnectionFailure(err) from err
        except httpx.HTTPError as err:
            raise GeneralFailure(err) from err

    async def _getBinary(self, path: str) -> Tuple[Optional[bytes], Optional[str]]:

        try:
            resp = await self.session.get(self._url(path))
            if resp.status_code == 401:
                raise AutenticationFailure("Authenticaion failed to device")

            if resp.status_code != 200:
                return None, None
            return resp.content, resp.headers.get("content-type")
        except (httpx.ConnectTimeout, httpx.ConnectError) as err:
            raise ConnectionFailure(err) from err
        except httpx.HTTPError as err:
            raise GeneralFailure(err) from err

    async def postReq(self, path: str, data: Any, timeout=None, protocol=None) -> Optional[Dict]:
        try:
            resp = await self.session.post(self._url(path, protocol), json=data, timeout=timeout)
            if resp.status_code == 401:
                raise AutenticationFailure("Authenticaion failed to device")

            if resp.status_code == 200:
                LOG.debug("Post succeded: %s -> %s", data, resp.text)
                return decode_xtv_response(resp)

            LOG.warning("Post failed: %s -> %s", data, resp.text)
            return None
        except httpx.ReadTimeout:
            LOG.debug("Read time out on postReq", exc_info=True)
            return None
        except (httpx.ConnectTimeout, httpx.ConnectError) as err:
            raise ConnectionFailure(err) from err
        except (httpx.ProtocolError, httpx.ReadError) as err:
            raise ProtocolFailure(err) from err
        except httpx.HTTPError as err:
            raise GeneralFailure(err) from err

    async def pairRequest(
        self,
        app_id: str,
        app_name: str,
        device_name: str,
        device_os: str,
        type: str,
        device_id: Optional[str] = None,
    ):
        """Start up a pairing request."""
        if device_id is None:
            device_id = token_hex(16)

        device = {
            "device_name": device_name,
            "device_os": device_os,
            "type": type,
            "id": device_id,
            "app_id": app_id,
            "app_name": app_name,
        }

        state = {"device": device}

        data = {
            "access": {
                "scope": ["read", "write", "control"]
            },
            "device": device
        }

        if self.system:
            featuring = self.system.get("featuring", None)
            if featuring:
                data["access"]["featuring"] = featuring


        LOG.debug("pair/request request: %s", data)
        resp = await self.session.post(self._url("pair/request"), json=data, auth=None)
        if not resp.headers["content-type"].startswith("application/json"):
            raise NoneJsonData(resp.text)
        data_response = resp.json()

        LOG.debug("pair/request response: %s", data_response)
        if data_response.get("error_id") != "SUCCESS":
            raise PairingFailure(data_response)

        state["timestamp"] = data_response["timestamp"]
        state["auth_key"] = data_response["auth_key"]

        return state

    async def pairGrant(self, state: Dict[str, Any], pin: str):
        """Finish a pairing sequence"""
        auth_handler = httpx.DigestAuth(state["device"]["id"], state["auth_key"])

        signature = hmac_signature(self.auth_shared_key, str(state["timestamp"]), pin)

        auth = {
            "auth_appId": "1",
            "auth_timestamp": state["timestamp"],
            "auth_signature": signature,
            "pin": pin,
        }

        data = {"auth": auth, "device": state["device"]}

        LOG.debug("pair/grant request: %s", data)
        resp = await self.session.post(
            self._url("pair/grant"), json=data, auth=auth_handler
        )
        if not resp.headers["content-type"].startswith("application/json"):
            raise NoneJsonData(resp.text)
        data_response = resp.json()
        LOG.debug("pair/grant response: %s", data_response)

        if data_response.get("error_id") != "SUCCESS":
            raise PairingFailure(data_response)

        self.session.auth = auth_handler
        return state["device"]["id"], state["auth_key"]

    async def setTransport(
        self,
        secured_transport: Optional[bool] = None,
        api_version: Optional[int] = None,
    ):
        if secured_transport == True and self.protocol != "https":
            LOG.info("Switching to secured transport")
            self.protocol = "https"
        elif secured_transport == False and self.protocol != "http":
            LOG.info("Switching to unsecured transport")
            self.protocol = "http"

        if api_version and api_version != self.api_version:
            LOG.info("Switching to api_version %d", api_version)
            self.api_version = api_version

    async def update(self):
        try:
            if not self.on:
                await self.getSystem()
                await self.setTransport(self.secured_transport)
                await self.getSources()
                await self.getChannelLists()
                await self.getChannels()
                await self.getApplications()
                await self.getAmbilightSupportedStyles()
                await self.getAmbilightCached()
                await self.getMenuItemsSettingsStructure()

            await self.getPowerState()
            await self.getAudiodata()
            await self.getSourceId()
            await self.getChannelId()
            await self.getApplication()
            await self.getContext()
            await self.getScreenState()
            await self.getAmbilightMode()
            await self.getAmbilightPower()
            await self.getAmbilightCurrentConfiguration()
            await self.getHueLampPower()
            await self.getRecordings()
            self.on = True
            return True
        except ConnectionFailure as err:
            LOG.debug("TV not available: %s", repr(err))
            self.on = False
            return False

    def _decodeSystem(self, system) -> SystemType:
        result = {}
        for key, value in system.items():
            if key.endswith("_encrypted"):
                result[key[:-10]] = cbc_decode(self.auth_shared_key, value)
            else:
                result[key] = value
        return cast(SystemType, result)

    async def getSystem(self):
        # Newest TV software requires https for system-info. Therefore we will try both protocols.
        protocols = ["http", "https"]

        if protocols[0] != self.protocol:
            protocols = reversed(protocols)
        
        for prot in protocols:
            r = cast(Optional[SystemType], await self.getReq("system", protocol=prot))
            if r:
                self.system = self._decodeSystem(r)
                self.name = r.get("name")
                return r
            else:
                self.system = {}
                self.name = None
        return r

    async def getAudiodata(self):
        r = await self.getReq("audio/volume")
        if r:
            self.audio_volume = r
        else:
            self.audio_volume = r
        return r

    async def getChannels(self):
        if self.api_version >= 5:
            self.channels = {}
            for channel_list in self.channel_lists.values():
                for channel in channel_list.get("Channel", []):
                    self.channels[str(channel.get("ccid"))] = channel
            return self.channels
        else:
            r = cast(Optional[ChannelsType], await self.getReq("channels"))
            if r:
                self.channels = r
            else:
                self.channels = {}
            return r

    async def getChannelId(self):
        if not self.channels:
            self.channel = None
            return None

        if self.api_version >= 5:
            r = cast(Optional[ActivitesTVType], await self.getReq("activities/tv"))
        else:
            r = cast(
                Optional[ChannelsCurrentType], await self.getReq("channels/current")
            )

        self.channel = r
        return r

    async def getChannelName(self, ccid) -> Optional[str]:
        if not self.channels:
            return None
        return self.channels.get(ccid, dict()).get("name", None)

    async def getChannelLogo(
        self, ccid, channel_list="all"
    ) -> Tuple[Optional[bytes], Optional[str]]:
        if self.api_version >= 5:
            data, content_type = await self._getBinary(
                f"channeldb/tv/channelLists/{channel_list}/{ccid}/logo"
            )
        else:
            data, content_type = await self._getBinary(f"channels/{ccid}/logo.png")
        return data, content_type

    async def getContext(self) -> Optional[ContextType]:
        r = cast(Optional[ContextType], await self.getReq(f"context"))
        self.context = r
        return r

    async def setChannel(self, ccid: Union[str, int], list_id: Optional[str] = None):
        channel: Union[ActivitesTVType, ChannelsCurrentType]
        if self.api_version >= 5:
            channel = {"channelList": {"id": list_id or "all"}, "channel": {"ccid": int(ccid)}}
            if await self.postReq("activities/tv", cast(Dict, channel)) is not None:
                self.channel = channel
                return True
        else:
            channel = {"id": str(ccid)}
            if await self.postReq("channels/current", cast(Dict, channel)) is not None:
                self.channel = channel
                return True
        return False

    async def getChannelLists(self):
        if self.api_version >= 5:
            r = cast(ChannelDbTv, await self.getReq("channeldb/tv"))
            if r:
                channel_lists = {}
                favorite_lists = {}

                for data in r.get("channelLists", []):
                    list_id = data["id"]
                    channel_list = cast(
                        Optional[ChannelListType],
                        await self.getReq(f"channeldb/tv/channelLists/{list_id}"),
                    )
                    if channel_list:
                        channel_lists[list_id] = channel_list


                for data in r.get("favoriteLists", []):
                    list_id = data["id"]
                    favorite_list = cast(
                        Optional[FavoriteListType],
                        await self.getReq(f"channeldb/tv/favoriteLists/{list_id}"),
                    )
                    if favorite_list:
                        favorite_lists[list_id] = favorite_list

                self.channel_lists = channel_lists
                self.favorite_lists = favorite_lists
            else:
                self.channel_lists = {}
                self.favorite_lists = {}

            return r

    async def getFavoriteList(self, list_id: str):
        if self.api_version >= 5:
            r = cast(
                Optional[FavoriteListType],
                await self.getReq(f"channeldb/tv/favoriteLists/{list_id}"),
            )
            if r and "channels" in r:
                return r["channels"]
            else:
                return None

    async def getChannelList(self, list_id: str):
        if self.api_version >= 5:
            r = cast(
                Optional[ChannelListType],
                await self.getReq(f"channeldb/tv/channelLists/{list_id}"),
            )
            if r and "Channel" in r:
                return r["Channel"]
            else:
                return None

    async def getSources(self):
        if self.json_feature_supported("activities", "intent"):
            self.sources = {
                channel_uri(None): {"name": "Watch TV"},
                passthrough_uri("com.mediatek.tvinput/.hdmi.HDMIInputService/HW5"): {
                    "name": "HDMI 1"
                },
                passthrough_uri("com.mediatek.tvinput/.hdmi.HDMIInputService/HW6"): {
                    "name": "HDMI 2"
                },
                passthrough_uri("com.mediatek.tvinput/.hdmi.HDMIInputService/HW7"): {
                    "name": "HDMI 3"
                },
                passthrough_uri("com.mediatek.tvinput/.hdmi.HDMIInputService/HW8"): {
                    "name": "HDMI 4"
                },
            }
            return self.sources

        if self.api_version == 1:
            r = await self.getReq("sources")
            if r:
                self.sources = r
            else:
                self.sources = {}
            return r

    async def getSourceId(self):
        if self.api_version < 5:
            r = await self.getReq("sources/current")
            if r:
                self.source_id = r["id"]
            else:
                self.source_id = None
            return r

    async def getSourceName(self, srcid) -> Optional[str]:
        if not self.sources:
            return None
        return self.sources.get(srcid, dict()).get("name", None)

    async def setSource(self, source_id):
        if self.api_version == 1:
            r = await self.postReq("sources/current", {"id": source_id})
            if r is not None:
                return True
            return False

        if self.json_feature_supported("activities", "intent"):
            if source_id == CHANNEL_URI and self.channel_id:
                source_id = channel_uri(self.channel_id)

            intent: ApplicationIntentType = {
                "extras": {"uri": source_id},
                "action": "org.droidtv.playtv.SELECTURI",
                "component": {
                    "packageName": "org.droidtv.playtv",
                    "className": "org.droidtv.playtv.PlayTvActivity",
                },
            }
            return await self.setApplication(intent)

        return False

    async def getApplications(self):
        if self.json_feature_supported("applications"):
            r = cast(ApplicationsType, await self.getReq("applications"))
            if r:
                self.applications = {app["id"]: app for app in r["applications"] if "id" in app}
            else:
                self.applications = {}
            return r

    async def getApplication(self):
        if self.json_feature_supported("applications"):
            r = cast(ApplicationIntentType, await self.getReq("activities/current"))
            if r:
                self.application = r
            else:
                self.application = None
            return r

    async def getApplicationIcon(self, id) -> Tuple[Optional[bytes], Optional[str]]:
        if self.json_feature_supported("applications"):
            data, content_type = await self._getBinary(f"applications/{id}/icon")
            return data, content_type
        else:
            return None, None

    async def getPowerState(self):
        r = await self.getReq("powerstate")
        if r:
            self.powerstate = cast(str, r["powerstate"])
        else:
            self.powerstate = None
        return r

    async def setPowerState(self, state):
        data = {"powerstate": state}
        if await self.postReq("powerstate", data) is not None:
            self.powerstate = state
            return True
        return False

    async def getScreenState(self):
        if self.api_version >= 5:
            r = await self.getReq("screenstate")
            if r:
                self.screenstate = cast(str, r["screenstate"])
            else:
                self.screenstate = None
            return r
        else:
            self.screenstate = None

    async def setScreenState(self, state):
        if self.api_version >= 5:
            data = {"screenstate": state}
            if await self.postReq("screenstate", data) is not None:
                self.screenstate = state
                return True
        return False

    async def getHueLampPower(self):
        if self.json_feature_supported("ambilight", "Hue"):
            r = await self.getReq("HueLamp/power")
            if r and "power" in r:
                self.huelamp_power = cast(str, r["power"])
            else:
                self.huelamp_power = None
            return r
        return None

    async def setHueLampPower(self, state):
        if self.json_feature_supported("ambilight", "Hue"):
            data = {"power": state}
            if await self.postReq("HueLamp/power", data) is not None:
                self.huelamp_power = state
                return True
        return False

    async def setApplication(self, intent: ApplicationIntentType):
        if self.json_feature_supported("activities", "intent"):
            data = {"intent": intent}
            if await self.postReq("activities/launch", data) is not None:
                self.application = intent
                return True
        return False

    async def setVolume(self, level, muted=False):
        data = {}
        if self.audio_volume is None or self.min_volume is None or self.max_volume is None:
            await self.getAudiodata()
        assert self.audio_volume is not None
        assert self.max_volume is not None
        assert self.min_volume is not None

        if level is not None:
            try:
                targetlevel = int(level * self.max_volume)
            except ValueError:
                LOG.warning("Invalid audio level %s" % str(level))
                return False
            if targetlevel < self.min_volume or targetlevel > self.max_volume:
                LOG.warning(
                    "Level not in range (%i - %i)" % (self.min_volume, self.max_volume)
                )
                return False
            data["current"] = targetlevel

        data["muted"] = muted

        if await self.postReq("audio/volume", data) is None:
            return False

        self.audio_volume.update(data)

    async def sendKey(self, key):
        res = await self.postReq("input/key", {"key": key}) is not None
        if res:
            if key == "Mute":
                if self.audio_volume:
                    self.audio_volume["muted"] = not self.audio_volume["muted"]
            elif key == "VolumeUp":
                if (
                    self.audio_volume
                    and self.audio_volume["current"] < self.audio_volume["max"]
                ):
                    self.audio_volume["current"] += 1
            elif key == "VolumeDown":
                if (
                    self.audio_volume
                    and self.audio_volume["current"] > self.audio_volume["min"]
                ):
                    self.audio_volume["current"] -= 1

    async def sendUnicode(self, key: str):
        return await self.postReq("input/key", {"unicode": key}) is not None

    async def getAmbilightMode(self):
        data = await self.getReq("ambilight/mode")
        if data:
            self.ambilight_mode_raw = data["current"]

            if self.quirk_ambilight_mode_ignored:
                # we could probably disable quirk here
                if data["current"] != "internal" and self.ambilight_mode_set:
                    LOG.warning(
                        "TV properly report ambilight mode, quirk should be disabled"
                    )
                    self.ambilight_mode_set = None

            return data["current"]

        else:
            self.ambilight_mode_raw = None

    async def setAmbilightMode(self, mode):
        data = {"current": mode}
        if await self.postReq("ambilight/mode", data) is None:
            return False

        if self.quirk_ambilight_mode_ignored:
            if mode == "internal":
                self.ambilight_mode_set = None
                if (
                    await self.postReq(
                        "ambilight/mode", {"current": "internal_invalid"}
                    )
                    is None
                ):
                    LOG.info("Ignoring error for workaround for ambilight mode")
            else:
                self.ambilight_mode_set = mode

        self.ambilight_mode_raw = mode

        return True

    async def getAmbilightTopology(self):
        r = await self.getReq("ambilight/topology")
        if r:
            self.ambilight_topology = r
        else:
            self.ambilight_topology = None
        return r

    async def getAmbilightMeasured(self):
        r = await self.getReq("ambilight/measured")
        if r:
            self.ambilight_measured = r
        else:
            self.ambilight_measured = None
        return r

    async def getAmbilightProcessed(self):
        r = await self.getReq("ambilight/processed")
        if r:
            self.ambilight_processed = r
        else:
            self.ambilight_processed = None
        return r

    async def getAmbilightCached(self):
        """Grab cached ambilight data from device. There are some unconfirmed reports where
        it seems the server crashes while grabbing this data from time to time. So use
        sparingly.
        """
        r = await self.getReq("ambilight/cached")
        if r:
            self.ambilight_cached = r
        else:
            self.ambilight_cached = None
        return r

    async def getAmbilightPower(self):
        if self.api_version >= 5:
            r = await self.getReq("ambilight/power")
            if r:
                self.ambilight_power_raw = r
            else:
                self.ambilight_power_raw = None
            return r

    async def setAmbilightPower(self, power: str):
        if self.api_version >= 5:
            data = {"power": power}
            if await self.postReq("ambilight/power", data) is None:
                return False
            self.ambilight_power_raw = data

            if self.quirk_ambilight_mode_ignored:
                if power == "On" and "lounge" in self.ambilight_modes:
                    self.ambilight_mode_raw = "lounge"
                    self.ambilight_mode_set = "lounge"

            return True

        if self.api_version == 1:
            if power == "On":
                await self.setAmbilightMode("internal")
            else:
                await self.setAmbilightMode("manual")
                await self.setAmbilightCached({"r": 0, "g": 0, "b": 0})

            return True

    async def setAmbilightCached(self, data):
        if await self.postReq("ambilight/cached", data) is None:
            return False

        # won't do optimistic behavior here since
        # the data format can differ a lot
        await self.getAmbilightCached()

        if self.quirk_ambilight_mode_ignored:
            if self.ambilight_mode_set not in ("manual", "expert"):
                self.ambilight_mode_raw = "manual"
                self.ambilight_mode_set = "manual"

        return True

    async def getAmbilightSupportedStyles(self):
        if self.json_feature_supported("ambilight", "Ambilight"):
            r = cast(
                Optional[AmbilightSupportedStylesType],
                await self.getReq("ambilight/supportedstyles"),
            )
            if r:
                self.ambilight_styles = {
                    style["styleName"]: style for style in r["supportedStyles"] if "styleName" in style
                }

                if self.quirk_ambilight_styles_menuitems:
                    style = self.ambilight_styles.setdefault(
                        "FOLLOW_VIDEO", {"styleName": "FOLLOW_VIDEO"}
                    )
                    style["menuSettings"] = [
                        "STANDARD",
                        "VIVID",
                        "IMMERSIVE",
                        "NATURAL",
                        "GAME",
                    ]

                    style = self.ambilight_styles.setdefault(
                        "FOLLOW_AUDIO", {"styleName": "FOLLOW_AUDIO"}
                    )
                    style["menuSettings"] = [
                        "ENERGY_ADAPTIVE_BRIGHTNESS",
                        "VU_METER",
                        "RANDOM_PIXEL_FLASH",
                    ]

                    if "Lounge light" in self.ambilight_styles:
                        style = self.ambilight_styles["Lounge light"]
                    elif "FOLLOW_COLOR" in self.ambilight_styles:
                        style = self.ambilight_styles["FOLLOW_COLOR"]
                    else:
                        style = self.ambilight_styles.setdefault(
                            "Lounge light", {"styleName": "Lounge light"}
                        )
                    style["menuSettings"] = [
                        "HOT_LAVA",
                        "DEEP_WATER",
                        "FRESH_NATURE",
                        "ISF",
                        "CUSTOM_COLOR",
                    ]
            else:
                self.ambilight_styles = {}
            return r

    async def getAmbilightCurrentConfiguration(self):
        if self.json_feature_supported("ambilight", "Ambilight"):
            r = cast(
                Optional[AmbilightCurrentConfiguration],
                await self.getReq("ambilight/currentconfiguration"),
            )
            if r:
                self.ambilight_current_configuration = r
            else:
                self.ambilight_current_configuration = None
            return r

    async def setAmbilightCurrentConfiguration(
        self, config: AmbilightCurrentConfiguration
    ):
        if self.json_feature_supported("ambilight", "Ambilight"):
            r = await self.postReq("ambilight/currentconfiguration", cast(Dict, config))
            if r is None:
                self.ambilight_current_configuration = None
                return False

            self.ambilight_current_configuration = config

            if self.quirk_ambilight_mode_ignored:
                self.ambilight_mode_set = None

            return True

    async def getRecordings(self):
        #Just known working with API level 6 currently
        if self.api_version == 6:
            if self.json_feature_supported("recordings", "List"):
                r = cast(
                    Optional[RecordingsListed],
                    await self.getReq("recordings/list"),
                )
                if r:
                    self.recordings_list = r
                else:
                    self.recordings_list = None
                return r

    async def openURL(self, url: str):
        if self.json_feature_supported("activities", "browser"):
            r = await self.postReq("activities/browser", {"url": url})
            return r is not None

    async def getStringsCached(self, string_ids: Iterable[str]) -> Union[Dict[str, str], None]:
        """Fetch translations for strings, making sure all items exist."""
        string_ids_wanted = set(string_ids)
        string_ids_missing = string_ids_wanted - self.strings.keys()

        # fetch translations from TV
        if string_ids_missing:
            if (data := await self.getStrings(sorted(string_ids_missing))) is None:
                return None
            self.strings.update(data)

        # make sure untranslated strings is set 1 to 1
        string_ids_missing = string_ids_wanted - set(self.strings.keys())
        for string_id in string_ids_missing:
            self.strings[string_id] = string_id

        return self.strings

    async def getStrings(
        self,
        strings: Iterable[str],
        language: Optional[str] = None,
        country: Optional[str] = None,
        variant: Optional[str] = None,
    ) -> Dict[str, str]:
        res: Dict[str, str] = {}
        for group in chunked_iterator(MAXIMUM_ITEMS_IN_REQUEST, strings):
            data = await self._getStrings(group, language=language, country=country, variant=variant)
            res.update(data)
        return res

    async def _getStrings(
        self,
        strings: Iterable[str],
        language: Optional[str] = None,
        country: Optional[str] = None,
        variant: Optional[str] = None,
    ) -> Dict[str, str]:
        data: StringsRequest = {
            "strings": [{"string_id": string} for string in strings],
            "locale": {
                "language": language or "",
                "country": country or "",
                "variant": variant or "",
            },
        }
        res = cast(Optional[Strings], await self.postReq("strings", data))
        if res:
            return {
                translation["string_id"]: translation["string_translation"]
                for translation in res["translations"]
                if translation["string_translation"]
            }
        return {}

    async def getMenuItemsSettingsStructure(self, force=False) -> Optional[MenuItemsSettingsStructure]:
        if self.json_feature_supported("menuitems", "Setup_Menu") or force:
            r = cast(Optional[MenuItemsSettingsStructure],
                    await self.getReq("menuitems/settings/structure")
            )
            self.settings = r
            return r

    async def getMenuItemsSettingsCurrent(self, node_ids: Iterable[int], force=False) -> Optional[MenuItemsSettingsCurrent]:
        current: MenuItemsSettingsCurrent = {"values": [], "version": 0}
        for group in chunked_iterator(MAXIMUM_ITEMS_IN_REQUEST, node_ids):
            if (data := await self._getMenuItemsSettingsCurrent(group)) is None:
                return None

            current["values"].extend(data["values"])
            if current["version"] == 0:
                current["version"] = data["version"]
        return current

    async def _getMenuItemsSettingsCurrent(self, node_ids: Iterable[int], force=False) -> Optional[MenuItemsSettingsCurrent]:
        if not node_ids:
            return None

        if self.json_feature_supported("menuitems", "Setup_Menu") or force:
            post: MenuItemsSettingsCurrentPost = {
                "nodes": [
                    {
                        "nodeid": node_id
                    }
                    for node_id in node_ids
                ] 
            }
            r = cast(Optional[MenuItemsSettingsCurrent],
                     await self.postReq("menuitems/settings/current", cast(dict, post))
            )
            return r

    async def getMenuItemsSettingsCurrentValue(self, node_ids: List[int], force=False) -> Dict[int, Optional[MenuItemsSettingsCurrentValueValue]]:
            data = await self.getMenuItemsSettingsCurrent(node_ids, force=force)
            res: Dict[int, Optional[MenuItemsSettingsCurrentValueValue]] = {}
            if data:
                for value in data.get("values", {}):
                    value_value = value["value"]

                    # Sometimes the API places the data field outside of the value field.
                    # This seem faulty, so let's move it in.
                    if data := value.get("data"):
                        value_value["data"] = data

                    res[value["value"].get("Nodeid")] = value_value
            for node_id in node_ids:
                if node_id not in res:
                    res[node_id] = None
            return res

    async def postMenuItemsSettingsUpdate(self, post: MenuItemsSettingsUpdate, force=False):
        if self.json_feature_supported("menuitems", "Setup_Menu") or force:
            return await self.postReq("menuitems/settings/update", cast(dict, post))

    async def postMenuItemsSettingsUpdateData(self, value: Dict[int, MenuItemsSettingsValueData], force=False):
        post = {
            "values": [
                {
                    "value": {
                        "Nodeid": node_id,
                        "data": data
                    },
                }
                for node_id, data in value.items()
            ]
        }
        return await self.postMenuItemsSettingsUpdate(cast(MenuItemsSettingsUpdate, post), force=force)
        

    async def notifyChange(self, timeout=TIMEOUT_NOTIFYREAD):
        """Start a http connection waiting for changes."""
        if not self.notify_change_supported:
            return None

        data = {
            "notification": {
                "activities/tv": self.channel,
                "activities/current": self.application,
                "powerstate": {"powerstate": self.powerstate},
                "audio/volume": self.audio_volume,
                "context": self.context,
                "screenstate": {"screenstate": self.screenstate},
                "menuitems/settings/version": {"version": self.settings_version}
            }
        }

        timeout_ctx = httpx.Timeout(timeout=TIMEOUT, connect=TIMEOUT_CONNECT, read=timeout)
        try:
            result = await self.postReq("notifychange", data=data, timeout=timeout_ctx)
        except ProtocolFailure as err:
            # not uncommon for tv to close connection on the lingering connection
            LOG.debug("Protocol failure from device: %s", repr(err))
            result = False
        except ConnectionFailure as err:
            LOG.debug("Connection failure to device: %s", repr(err))
            self.on = False
            result = None

        if result:
            if "activities/tv" in result:
                self.channel = result["activities/tv"]

            if "activities/current" in result:
                self.application = result["activities/current"]

            if "powerstate" in result:
                self.powerstate = result["powerstate"]["powerstate"]

            if "audio/volume" in result:
                self.audio_volume = result["audio/volume"]

            if "context" in result:
                self.context = result["context"]

            if "screenstate" in result:
                self.screenstate = result["screenstate"]["screenstate"]

            if "menuitems/settings/version" in result:
                self.settings_version = result["menuitems/settings/version"]["version"]

            return True

        if result is None:
            return None

        return False
